# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
import json
import tempfile
import types

import numpy as np
import os
import six
import re
import networkx as nx
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras import initializers
from tensorflow.python.keras.layers.core import TFOpLambda

from tensorflow.keras.models import Model
from tensorflow.keras.models import model_from_json

from tensorflow_model_optimization.python.core.sparsity.keras import pruning_wrapper
from tensorflow_model_optimization.python.core.sparsity.keras import prune_registry
from tensorflow_model_optimization.python.core.sparsity.keras import prunable_layer

from .qlayers import Clip
from .qconv2d_batchnorm import QConv2DBatchnorm
from .qdepthwiseconv2d_batchnorm import QDepthwiseConv2DBatchnorm
from .qlayers import QActivation
from .qlayers import QAdaptiveActivation
from .qpooling import QAveragePooling2D
from .qlayers import QDense
from .qlayers import QInitializer
from .qconvolutional import QConv1D
from .qconvolutional import QConv2D
from .qconvolutional import QConv2DTranspose
from .qrecurrent import QSimpleRNN
from .qrecurrent import QSimpleRNNCell
from .qrecurrent import QLSTM
from .qrecurrent import QLSTMCell
from .qrecurrent import QGRU
from .qrecurrent import QGRUCell
from .qrecurrent import QBidirectional
from .qconvolutional import QSeparableConv1D
from .qconvolutional import QSeparableConv2D
from .qconvolutional import QDepthwiseConv2D
from .qnormalization import QBatchNormalization
from .qpooling import QGlobalAveragePooling2D
from .qtools import qgraph
from .quantizers import binary
from .quantizers import bernoulli
from .quantizers import get_weight_scale
from .quantizers import quantized_bits
from .quantizers import quantized_relu
from .quantizers import quantized_ulaw
from .quantizers import quantized_tanh
from .quantizers import quantized_sigmoid
from .quantizers import quantized_po2
from .quantizers import quantized_relu_po2
from .quantizers import stochastic_binary
from .quantizers import stochastic_ternary
from .quantizers import ternary

from .safe_eval import safe_eval

REGISTERED_LAYERS = [
    "QActivation",
    "QAdaptiveActivation",
    "QDense",
    "QConv1D",
    "QConv2D",
    "QSeparableConv1D",
    "QSeparableConv2D",
    "QDepthwiseConv2D",
    "QConv2DTranspose",
    "QSimpleRNN",
    "QLSTM",
    "QGRU",
    "QBidirectional",
    "QBatchNormalization",
    "QConv2DBatchnorm",
    "QDepthwiseConv2DBatchnorm",
    "QAveragePooling2D",
    "QGlobalAveragePooling2D",
]


#
# Model utilities: before saving the weights, we want to apply the quantizers
#
def model_save_quantized_weights(model, filename=None):
  """Quantizes model for inference and save it.

  Takes a model with weights, apply quantization function to weights and
  returns a dictionary with quantized weights.

  User should be aware that "po2" quantization functions cannot really
  be quantized in meaningful way in Keras. So, in order to preserve
  compatibility with inference flow in Keras, we do not covert "po2"
  weights and biases to exponents + signs (in case of quantize_po2), but
  return instead (-1)**sign*(2**round(log2(x))). In the returned dictionary,
  we will return the pair (sign, round(log2(x))).

  Arguments:
    model: model with weights to be quantized.
    filename: if specified, we will save the hdf5 containing the quantized
      weights so that we can use them for inference later on.

  Returns:
    dictionary containing layer name and quantized weights that can be used
    by a hardware generator.

  """

  saved_weights = {}

  print("... quantizing model")
  for layer in model.layers:
    if hasattr(layer, "get_quantizers"):
      weights = []
      signs = []

      if any(isinstance(layer, t) for t in [
          QConv2DBatchnorm, QDepthwiseConv2DBatchnorm]):
        qs = layer.get_quantizers()
        ws = layer.get_folded_weights()
      elif any(isinstance(layer, t) for t in [QSimpleRNN, QLSTM, QGRU]):
        qs = layer.get_quantizers()[:-1]
        ws = layer.get_weights()
      else:
        qs = layer.get_quantizers()
        ws = layer.get_weights()

      has_sign = False

      for quantizer, weight in zip(qs, ws):
        if quantizer:
          weight = tf.constant(weight)
          weight = tf.keras.backend.eval(quantizer(weight))

        # If quantizer is power-of-2 (quantized_po2 or quantized_relu_po2),
        # we would like to process it here.
        #
        # However, we cannot, because we will loose sign information as
        # quanized_po2 will be represented by the tuple (sign, log2(abs(w))).
        #
        # In addition, we will not be able to use the weights on the model
        # any longer.
        #
        # So, instead of "saving" the weights in the model, we will return
        # a dictionary so that the proper values can be propagated.

        weights.append(weight)

        has_sign = False
        if quantizer:
          if isinstance(quantizer, six.string_types):
            q_name = quantizer
          elif hasattr(quantizer, "__name__"):
            q_name = quantizer.__name__
          elif hasattr(quantizer, "name"):
            q_name = quantizer.name
          elif hasattr(quantizer, "__class__"):
            q_name = quantizer.__class__.__name__
          else:
            q_name = ""
        if quantizer and ("_po2" in q_name):
          # Quantized_relu_po2 does not have a sign
          if isinstance(quantizer, quantized_po2):
            has_sign = True
          sign = np.sign(weight)
          # Makes sure values are -1 or +1 only
          sign += (1.0 - np.abs(sign))
          weight = np.round(np.log2(np.abs(weight)))
          signs.append(sign)
        else:
          signs.append([])

      saved_weights[layer.name] = {"weights": weights}
      if has_sign:
        saved_weights[layer.name]["signs"] = signs

      if not any(isinstance(layer, t) for t in [
          QConv2DBatchnorm, QDepthwiseConv2DBatchnorm]):
        layer.set_weights(weights)
      else:
        print(layer.name, " conv and batchnorm weights cannot be seperately"
              " quantized because they will be folded before quantization.")
    else:
      if layer.get_weights():
        print(" ", layer.name, "has not been quantized")

  if filename:
    model.save_weights(filename)

  return saved_weights


def quantize_activation(layer_config, activation_bits):
  """Replaces activation by quantized activation functions."""
  str_act_bits = str(activation_bits)
  # relu -> quantized_relu(bits)
  # tanh -> quantized_tanh(bits)
  # sigmoid -> quantized_sigmoid(bits)
  # more to come later
  if layer_config.get("activation", None) is None:
    return
  if isinstance(layer_config["activation"], six.string_types):
    a_name = layer_config["activation"]
  elif isinstance(layer_config["activation"], types.FunctionType):
    a_name = layer_config["activation"].__name__
  else:
    a_name = layer_config["activation"].__class__.__name__

  if a_name == "linear":
    return
  if a_name == "relu":
    layer_config["activation"] = "quantized_relu(" + str_act_bits + ")"
  elif a_name == "tanh":
    layer_config["activation"] = "quantized_tanh(" + str_act_bits + ")"
  elif a_name == "sigmoid":
    layer_config["activation"] = "quantized_sigmoid(" + str_act_bits + ")"


def get_config(quantizer_config, layer, layer_class, parameter=None):
  """Returns search of quantizer on quantizer_config."""
  quantizer = quantizer_config.get(layer["config"]["name"],
                                   quantizer_config.get(layer_class, None))

  if quantizer is not None and parameter is not None:
    quantizer = quantizer.get(parameter, None)

  return quantizer


def is_TFOpLambda_layer(layer):
  return (isinstance(layer, TFOpLambda) or
          layer.__class__.__name__ == "TFOpLambda")


def get_y_from_TFOpLambda(model_cfg, layer):
  """Get the value of "y" from the TFOpLambda layer's configuration.
  Args:
    model_cfg: dictionary type, model.get_config() output
    layer: a given layer instance

  Return:
    value of "y" for a TFOpLambda layer. 'y' here corresponds to how tensorflow
    stores TFOpLambda layer parameter in serialization. for example,
    TFOpLambda(func), where func is tf.multiply(input_tensor, 3). "y" would be
    the value 3.
  """

  for layer_config in model_cfg["layers"]:
    op_name = layer_config["config"]["name"]
    class_name = layer_config["class_name"]

    # TODO(lishanok): Extend support for other TFOpLambda types when needed
    if op_name == layer.name and  class_name == "TFOpLambda":
      assert ("tf.__operators__.add" in op_name or "tf.math.multiply"
              in op_name), "TFOpLambda layer {} not supported!".format(op_name)
      return layer_config["inbound_nodes"][-1][-1]["y"]

  return None


def convert_to_folded_model(model):
  """Find conv/dense layers followed by bn layers and fold them.

  Args:
    model: input model

  Returns:
    new model without bn layers
    list of layers being folded

  Note: supports sequential and non-sequential model
  """

  fold_model = clone_model(model)
  model_cfg = model.get_config()
  (graph, _) = qgraph.GenerateGraphFromModel(
      fold_model, "quantized_bits(8, 0, 1)", "quantized_bits(8, 0, 1)")

  qgraph.GraphAddSingleSourceSingleSink(graph)
  qgraph.GraphRemoveNodeWithNodeType(graph, "InputLayer")
  qgraph.GraphPropagateActivationsToEdges(graph)

  # Finds the Batchnorm nodes to be deleted and mark them.
  bn_nodes_to_delete = []
  layers_to_fold = []
  for node_id in nx.topological_sort(graph):
    layer_input_tensors = []
    node = graph.nodes[node_id]
    layer = node["layer"][0]
    if layer:
      successor_ids = list(graph.successors(node_id))
      is_single = len(successor_ids) == 1
      successor_layer = graph.nodes[successor_ids[0]]["layer"][0]
      followed_by_bn = (successor_layer.__class__.__name__ ==
                        "BatchNormalization")
      # TODO(lishanok): extend to QDense types
      is_foldable = layer.__class__.__name__ in [
          "Conv2D", "DepthwiseConv2D"
      ] and is_single and followed_by_bn

      if is_foldable:
        # Removes the batchnorm node from the graph.
        bn_nodes_to_delete.append(successor_ids[0])
        layers_to_fold.append(layer.name)

  # Deletes the marked nodes.
  for node_id in bn_nodes_to_delete:
    qgraph.GraphRemoveNode(graph, node_id)

  # Modifies model according to the graph.
  model_outputs = []
  x = model_inputs = fold_model.inputs

  for node_id in nx.topological_sort(graph):
    layer_input_tensors = []
    node = graph.nodes[node_id]

    layer = node["layer"][0]
    if layer:
      # Gets layer input tensors from graph edge.
      for parent_node_id in graph.predecessors(node_id):
        edge = graph.edges[(parent_node_id, node_id)]
        input_tensor = edge["tensor"]
        layer_input_tensors.append(input_tensor)

      # We call the layer to get output tensor.
      if len(layer_input_tensors) == 1:
        layer_input_tensors = layer_input_tensors[0].deref()
      else:
        layer_input_tensors = [t.deref() for t in layer_input_tensors]

      if is_TFOpLambda_layer(layer):
        # TFOpLambda layer requires one extra input: "y"
        y = get_y_from_TFOpLambda(model_cfg, layer)
        x = layer(layer_input_tensors, y)
      else:
        x = layer(layer_input_tensors)

      # Replaces edge tensors between the predecessor and successor
      for u, v in graph.edges(node_id):
        # u is current layer node, v is successor layer node
        # graph[u][v] is the edge between the two nodes
        # Replace the tensor on this edge so that the input tensor for the
        # successor layer can be updated accordingly.
        graph[u][v]["tensor"] = x.ref()

        if v == -2 and x not in model_outputs:
          # When it is output layer, add the output tensor of this layer
          # into model outputs.
          model_outputs.append(x)

  new_model = Model(inputs=model_inputs, outputs=model_outputs)

  return new_model, layers_to_fold


def model_quantize(model,
                   quantizer_config,
                   activation_bits,
                   custom_objects=None,
                   transfer_weights=False,
                   prefer_qadaptiveactivation=False,
                   enable_bn_folding=False):
  """Creates a quantized model from non-quantized model.

  The quantized model translation is based on json interface of Keras,
  which requires a custom_objects dictionary for "string" types.

  Because of the way json works, we pass "string" objects for the
  quantization mechanisms and we perform an eval("string") which
  technically is not safe, but it will do the job.

  The quantizer_config is a dictionary with the following form.
  {
    Dense_layer_name: {
        "kernel_quantizer": "quantizer string",
        "bias_quantizer": "quantizer_string"
    },

    Conv2D_layer_name: {
        "kernel_quantizer": "quantizer string",
        "bias_quantizer": "quantizer_string"
    },

    Activation_layer_name: "quantizer string",

    "QActivation": { "relu": "quantizer_string" },

    "QConv2D": {
        "kernel_quantizer": "quantizer string",
        "bias_quantizer": "quantizer_string"
    },

    "QBatchNormalization": {}
  }

  In the case of "QBidirectional", we can follow the same form as above.
  The specified configuration will be used for both forward and backwards
  layer.
  {
    "Bidirectional" : {
        "kernel_quantizer" : "quantizer string",
        "bias_quantizer" : "quantizer string",
        "recurrent_quantizer" : "quantizer string"
    }
  }

  In the case of "QActivation", we can modify only certain types of
  activations, for example, a "relu". In this case we represent the
  activation name by a dictionary, or we can modify all activations,
  without representhing as a set.

  We right now require a default case in case we cannot find layer name.
  This simplifies the dictionary because the simplest case, we can just
  say:

  {
    "default": {
        "kernel": "quantized_bits(4)",
        "bias": "quantized_bits(4)"
    }
  }

  and this will quantize all layers' weights and bias to be created with
  4 bits.

  Arguments:
    model: model to be quantized
    quantizer_config: dictionary (as above) with quantized parameters
    activation_bits: number of bits for quantized_relu, quantized_tanh,
      quantized_sigmoid
    custom_objects: dictionary following keras recommendations for json
      translation.
    transfer_weights: if true, weights are to be transfered from model to
      qmodel.
    prefer_qadaptiveactivation: Bool. If true, try to use QAdaptiveActivation
      over QActivation whenever possible
    enable_bn_folding: Bool. If true, fold conv/dense layers with
      following batch normalization layers whenever possible. use
      QConv2DBatchnorm for example, to replace conv2d layers

  Returns:
    qmodel with quantized operations and custom_objects.
  """

  if enable_bn_folding:
    # Removes bn layers from the model and find a list of layers to fold.
    model, layers_to_fold = convert_to_folded_model(model)
    if len(layers_to_fold) == 0:
      # If no layers to fold, no need to perform folding.
      enable_bn_folding = False

  if not custom_objects:
    custom_objects = {}

  # Let's make a deep copy to make sure our objects are not shared elsewhere.
  jm = copy.deepcopy(json.loads(model.to_json()))
  custom_objects = copy.deepcopy(custom_objects)
  config = jm["config"]
  layers = config["layers"]

  def quantize_rnn(layer, quantizer_config):
    q_name = "Q" + layer["class_name"]
    # Needs to add kernel, recurrent bias quantizers.
    kernel_quantizer = get_config(
        quantizer_config, layer, q_name, "kernel_quantizer")
    recurrent_quantizer = get_config(
        quantizer_config, layer, q_name, "recurrent_quantizer")
    if layer["config"]['use_bias']:
      bias_quantizer = get_config(
          quantizer_config, layer, q_name, "bias_quantizer")
    else:
      bias_quantizer = None
    state_quantizer = get_config(
            quantizer_config, layer, q_name, "state_quantizer")

    # This is to avoid unwanted transformations.
    if kernel_quantizer is None:
      return

    layer["config"]["kernel_quantizer"] = kernel_quantizer
    layer["config"]["recurrent_quantizer"] = recurrent_quantizer
    layer["config"]["bias_quantizer"] = bias_quantizer
    layer["config"]["state_quantizer"] = state_quantizer

    # If activation is present, add activation here.
    activation = get_config(
        quantizer_config, layer, q_name, "activation_quantizer")
    if activation:
      layer["config"]["activation"] = activation
    else:
      quantize_activation(layer["config"], activation_bits)

    # If recurrent activation is present, add activation here.
    if layer["class_name"] in ["LSTM", "GRU"]:
      recurrent_activation = get_config(
          quantizer_config, layer, q_name, "recurrent_activation_quantizer")
      if recurrent_activation:
        layer["config"]["recurrent_activation"] = recurrent_activation
    layer["class_name"] = q_name

  for layer in layers:
    layer_config = layer["config"]

    # Dense becomes QDense, Conv1D becomes QConv1D etc
    # Activation converts activation functions.

    if layer["class_name"] in [
      "Dense", "Conv1D", "Conv2D", "Conv2DTranspose",
      "SeparableConv1D", "SeparableConv2D"
    ]:
      if (layer["class_name"] in ["Dense", "Conv2D"] and enable_bn_folding and
          layer["name"] in layers_to_fold):
        # Only fold if current layer is followed by BN layer.
        q_name = "Q" + layer["class_name"] + "Batchnorm"
        layer_config["use_bias"] = True  # Folded layers require a bias

        # Sets ema_freeze_delay and folding_mode specific to
        # QDepthwiseConv2DBatchnorm layer config.
        folding_mode = get_config(
            quantizer_config, layer, q_name, "folding_mode")
        layer_config["folding_mode"] = (
            folding_mode if folding_mode else "ema_stats_folding")
        ema_freeze_delay = get_config(
            quantizer_config, layer, q_name, "ema_freeze_delay")
        layer_config["ema_freeze_delay"] = (
            ema_freeze_delay if ema_freeze_delay else None)
      else:
        q_name = "Q" + layer["class_name"]
      # Needs to add kernel/bias quantizers.
      kernel_quantizer = get_config(
          quantizer_config, layer, q_name, "kernel_quantizer")

      if layer_config["use_bias"]:
        bias_quantizer = get_config(
            quantizer_config, layer, q_name, "bias_quantizer")
      else:
        bias_quantizer = None

      if (kernel_quantizer is None and
          q_name == "Q" + layer["class_name"] + "Batchnorm"):
        # Tries none-folded layer quantizer as a back up.
        kernel_quantizer = get_config(
            quantizer_config, layer, "Q" + layer["class_name"],
            "kernel_quantizer")
        bias_quantizer = get_config(
            quantizer_config, layer, "Q" + layer["class_name"],
            "bias_quantizer")

      # This is to avoid unwanted transformations.
      if kernel_quantizer is None:
        continue

      layer["class_name"] = q_name

      layer_config["kernel_quantizer"] = kernel_quantizer
      layer_config["bias_quantizer"] = bias_quantizer

      # If activation is present, add activation here.
      quantizer = get_config(
          quantizer_config, layer, q_name, "activation_quantizer")

      if quantizer:
        layer_config["activation"] = quantizer
      else:
        quantize_activation(layer_config, activation_bits)

    elif layer["class_name"] == "DepthwiseConv2D":
      if enable_bn_folding and layer["name"] in layers_to_fold:
        q_name = "QDepthwiseConv2DBatchnorm"
        layer_config["use_bias"] = True  # Folded layers require a bias

        # Sets ema_freeze_delay and folding_mode specific to
        # QDepthwiseConv2DBatchnorm layers.
        folding_mode = get_config(
            quantizer_config, layer, q_name, "folding_mode")
        layer_config["folding_mode"] = (
            folding_mode if folding_mode else "ema_stats_folding")
        ema_freeze_delay = get_config(
            quantizer_config, layer, q_name, "ema_freeze_delay")
        layer_config["ema_freeze_delay"] = (
            ema_freeze_delay if ema_freeze_delay else None)

      else:
        q_name = "QDepthwiseConv2D"

      # Needs to add kernel/bias quantizers.
      depthwise_quantizer = get_config(quantizer_config, layer, q_name,
                                       "depthwise_quantizer")

      if layer_config["use_bias"]:
        bias_quantizer = get_config(quantizer_config, layer, q_name,
                                    "bias_quantizer")
      else:
        bias_quantizer = None

      if depthwise_quantizer is None and q_name == "QDepthwiseConv2DBatchnorm":
        # Tries none-folded layer quantizer as a back up.
        depthwise_quantizer = get_config(
            quantizer_config, layer, "QDepthwiseConv2D", "depthwise_quantizer")
        bias_quantizer = get_config(
            quantizer_config, layer, "QDepthwiseConv2D", "bias_quantizer")

      # This is to avoid unwanted transformations.
      if depthwise_quantizer is None:
        continue

      layer["class_name"] = q_name

      layer_config["depthwise_quantizer"] = depthwise_quantizer
      layer_config["bias_quantizer"] = bias_quantizer
      # If activation is present, add activation here.
      quantizer = get_config(quantizer_config, layer, q_name,
                             "activation_quantizer",)

      if quantizer:
        layer_config["activation"] = quantizer
      else:
        quantize_activation(layer_config, activation_bits)

    elif layer["class_name"] in ["SimpleRNN", "LSTM", "GRU"]:
      quantize_rnn(layer, quantizer_config)

    elif layer["class_name"] == "Bidirectional":
      forward_layer_quantizer_config = {
          layer_config["layer"]["config"]["name"]:
              get_config(quantizer_config, layer, "QBidirectional")
      }
      quantize_rnn(layer["config"]["layer"], forward_layer_quantizer_config)
      if "backward_layer" in layer_config:
        backward_layer_quantizer_config = {
            layer_config["backward_layer"]["config"]["name"]:
                get_config(quantizer_config, layer, "QBidirectional")
        }
        quantize_rnn(layer["config"]["backward_layer"],
                     backward_layer_quantizer_config)
      layer["class_name"] = "QBidirectional"

    elif layer["class_name"] == "Activation":
      if prefer_qadaptiveactivation:  # Try to find QAdaptiveActivation first
        quantizer = get_config(quantizer_config, layer, "QAdaptiveActivation")
        is_qadaptiveactivation = True
        if quantizer is None:  # Try QActivation as a backup
          quantizer = get_config(quantizer_config, layer, "QActivation")
          is_qadaptiveactivation = False
      else:  # Tries to find QActivation first.
        quantizer = get_config(quantizer_config, layer, "QActivation")
        is_qadaptiveactivation = False
        if quantizer is None:  # Try QAdaptiveActivation as a backup
          quantizer = get_config(quantizer_config, layer, "QAdaptiveActivation")
          is_qadaptiveactivation = True

      # This is to avoid softmax from quantizing in autoq.
      if quantizer is None:
        continue

      # If quantizer exists in dictionary related to this name,
      # use it, otherwise, use normal transformations.

      if not isinstance(quantizer, dict) or quantizer.get(
          layer_config["activation"], None):
        # Only change activation layer if we will use a quantized activation.

        layer["class_name"] = ("QAdaptiveActivation" if is_qadaptiveactivation
                               else "QActivation")
        if isinstance(quantizer, dict):
          quantizer = quantizer[layer_config["activation"]]
        if quantizer:
          if is_qadaptiveactivation:
            assert quantizer.find(",") < 0, \
                "Only integer bits should be defined for QAdaptiveActivation"
            layer_config["total_bits"] = int(re.sub(r"[^\d]", "", quantizer))
            quantizer = re.sub(r"\(.*", "", quantizer)  # remove params
          layer_config["activation"] = quantizer
        else:
          quantize_activation(layer_config, activation_bits)

    # We have to do this because of other instances of ReLU.
    elif layer["class_name"] in ["ReLU", "relu", "LeakyReLU"]:

      quantizer = get_config(quantizer_config, layer, "QActivation")
      # This is to avoid unwanted transformations.
      if quantizer is None:
        continue

      if layer["class_name"] == "LeakyReLU":
        negative_slope = layer["config"]["alpha"]
      elif layer["class_name"] == "relu":
        max_value = layer["config"]["max_value"]
        negative_slope = layer["config"]["alpha"]
        threshold = layer["config"]["threshold"]
      else: # ReLU from mobilenet
        max_value = layer["config"]["max_value"]
        negative_slope = layer["config"]["negative_slope"]
        threshold = layer["config"]["threshold"]

      if negative_slope > 0:
        q_name = "leakyrelu"
      else:
        q_name = "relu"

      # If quantizer exists in dictionary related to this name,
      # use it, otherwise, use normal transformations.

      if not isinstance(quantizer, dict) or quantizer.get(q_name, None):
        # Only change activation layer if we will use a quantized activation.

        layer["class_name"] = "QActivation"

        # Remove relu specific configurations
        # remember that quantized relu's are always upper bounded.

        if layer["class_name"] == "LeakyReLU":
          del layer["config"]["alpha"]
        elif layer["class_name"] == "relu":
          del layer["config"]["max_value"]
          del layer["config"]["alpha"]
          del layer["config"]["threshold"]
        else: # ReLU from mobilenet
          del layer["config"]["max_value"]
          del layer["config"]["negative_slope"]
          del layer["config"]["threshold"]

        if isinstance(quantizer, dict):
          quantizer = quantizer[q_name]
        if quantizer:
          layer["config"]["activation"] = quantizer
        else:
          quantize_activation(layer["config"], activation_bits)

    elif layer["class_name"] == "BatchNormalization":
      # We will assume at least QBatchNormalization or
      # layer name is in dictionary to enable conversion
      # otherwise we will just skip it.
      if (
          layer_config["name"] not in quantizer_config and
          "QBatchNormalization" not in quantizer_config
      ):
        continue

      layer["class_name"] = "QBatchNormalization"
      # Needs to add kernel/bias quantizers.
      gamma_quantizer = get_config(
          quantizer_config, layer, "QBatchNormalization",
          "gamma_quantizer")
      beta_quantizer = get_config(
          quantizer_config, layer, "QBatchNormalization",
          "beta_quantizer")
      mean_quantizer = get_config(
          quantizer_config, layer, "QBatchNormalization",
          "mean_quantizer")
      variance_quantizer = get_config(
          quantizer_config, layer, "QBatchNormalization",
          "variance_quantizer")

      layer_config["gamma_quantizer"] = gamma_quantizer
      layer_config["beta_quantizer"] = beta_quantizer
      layer_config["mean_quantizer"] = mean_quantizer
      layer_config["variance_quantizer"] = variance_quantizer

    elif layer["class_name"] in ["AveragePooling2D", "GlobalAveragePooling2D"]:
      q_name = "Q" + layer["class_name"]
      # Adds the average quanizer to config.
      average_quantizer = get_config(
          quantizer_config, layer, q_name, "average_quantizer")

      # This is to avoid unwanted transformations.
      if average_quantizer is None:
        continue

      layer["class_name"] = q_name

      layer_config["average_quantizer"] = average_quantizer

      # Adds activation to config.
      quantizer = get_config(
          quantizer_config, layer, q_name, "activation_quantizer")

      if quantizer:
        layer_config["activation"] = quantizer
      else:
        quantize_activation(layer_config, activation_bits)

  # We need to keep a dictionary of custom objects as our quantized library
  # is not recognized by keras.

  qmodel = quantized_model_from_json(json.dumps(jm), custom_objects)

  # If transfer_weights is true, we load the weights from model to qmodel.

  if transfer_weights and not enable_bn_folding:
    for layer, qlayer in zip(model.layers, qmodel.layers):
      if layer.get_weights():
        qlayer.set_weights(copy.deepcopy(layer.get_weights()))

  return qmodel


def _add_supported_quantized_objects(custom_objects):
  """Map all the quantized objects."""
  custom_objects["QInitializer"] = QInitializer
  custom_objects["QDense"] = QDense
  custom_objects["QConv1D"] = QConv1D
  custom_objects["QConv2D"] = QConv2D
  custom_objects["QConv2DTranspose"] = QConv2DTranspose
  custom_objects["QSimpleRNNCell"] = QSimpleRNNCell
  custom_objects["QSimpleRNN"] = QSimpleRNN
  custom_objects["QLSTMCell"] = QLSTMCell
  custom_objects["QLSTM"] = QLSTM
  custom_objects["QGRUCell"] = QGRUCell
  custom_objects["QGRU"] = QGRU
  custom_objects["QBidirectional"] = QBidirectional
  custom_objects["QDepthwiseConv2D"] = QDepthwiseConv2D
  custom_objects["QSeparableConv1D"] = QSeparableConv1D
  custom_objects["QSeparableConv2D"] = QSeparableConv2D
  custom_objects["QActivation"] = QActivation
  custom_objects["QAdaptiveActivation"] = QAdaptiveActivation
  custom_objects["QBatchNormalization"] = QBatchNormalization
  custom_objects["Clip"] = Clip
  custom_objects["quantized_bits"] = quantized_bits
  custom_objects["bernoulli"] = bernoulli
  custom_objects["stochastic_ternary"] = stochastic_ternary
  custom_objects["ternary"] = ternary
  custom_objects["stochastic_binary"] = stochastic_binary
  custom_objects["binary"] = binary
  custom_objects["quantized_relu"] = quantized_relu
  custom_objects["quantized_ulaw"] = quantized_ulaw
  custom_objects["quantized_tanh"] = quantized_tanh
  custom_objects["quantized_sigmoid"] = quantized_sigmoid
  custom_objects["quantized_po2"] = quantized_po2
  custom_objects["quantized_relu_po2"] = quantized_relu_po2
  

  custom_objects["QConv2DBatchnorm"] = QConv2DBatchnorm
  custom_objects["QDepthwiseConv2DBatchnorm"] = QDepthwiseConv2DBatchnorm

  custom_objects["QAveragePooling2D"] = QAveragePooling2D
  custom_objects["QGlobalAveragePooling2D"] = QGlobalAveragePooling2D


def clone_model(model, custom_objects=None):
  """Clones model with custom_objects."""
  if not custom_objects:
    custom_objects = {}

  # Makes a deep copy to make sure our objects are not shared elsewhere.
  custom_objects = copy.deepcopy(custom_objects)

  _add_supported_quantized_objects(custom_objects)

  json_string = model.to_json()
  qmodel = quantized_model_from_json(json_string, custom_objects=custom_objects)
  qmodel.set_weights(model.get_weights())

  return qmodel

  config = {
      "class_name": model.__class__.__name__,
      "config": model.get_config(),
  }
  clone = tf.keras.models.model_from_config(
      config, custom_objects=custom_objects)
  clone.set_weights(model.get_weights())
  return clone


def quantized_model_from_json(json_string, custom_objects=None):
  if not custom_objects:
    custom_objects = {}

  # Makes a deep copy to make sure our objects are not shared elsewhere.
  custom_objects = copy.deepcopy(custom_objects)

  _add_supported_quantized_objects(custom_objects)

  qmodel = model_from_json(json_string, custom_objects=custom_objects)

  return qmodel


def load_qmodel(filepath, custom_objects=None, compile=True):
  """Loads quantized model from Keras's model.save() h5 file.

  Arguments:
      filepath: one of the following:
          - string, path to the saved model
          - h5py.File or h5py.Group object from which to load the model
          - any file-like object implementing the method `read` that returns
          `bytes` data (e.g. `io.BytesIO`) that represents a valid h5py file
          image.
      custom_objects: Optional dictionary mapping names (strings) to custom
          classes or functions to be considered during deserialization.
      compile: Boolean, whether to compile the model after loading.

  Returns:
      A Keras model instance. If an optimizer was found as part of the saved
      model, the model is already compiled. Otherwise, the model is uncompiled
      and a warning will be displayed. When `compile` is set to False, the
      compilation is omitted without any warning.
  """

  if not custom_objects:
    custom_objects = {}

  # Makes a deep copy to make sure our objects are not shared elsewhere.
  custom_objects = copy.deepcopy(custom_objects)

  _add_supported_quantized_objects(custom_objects)

  qmodel = tf.keras.models.load_model(filepath, custom_objects=custom_objects,
                                      compile=compile)
  return qmodel


def print_model_sparsity(model):
  """Prints sparsity for the pruned layers in the model."""

  def _get_sparsity(weights):
    return 1.0 - np.count_nonzero(weights) / float(weights.size)

  print("Model Sparsity Summary ({})".format(model.name))
  print("--")
  for layer in model.layers:
    if isinstance(layer, pruning_wrapper.PruneLowMagnitude):
      prunable_weights = layer.layer.get_prunable_weights()
    elif isinstance(layer, prunable_layer.PrunableLayer):
      prunable_weights = layer.get_prunable_weights()
    elif prune_registry.PruneRegistry.supports(layer):
      weight_names = prune_registry.PruneRegistry._weight_names(layer)
      prunable_weights = [getattr(layer, weight) for weight in weight_names]
    else:
      prunable_weights = None
    if prunable_weights:
      print("{}: {}".format(
          layer.name, ", ".join([
              "({}, {})".format(weight.name,
                  str(_get_sparsity(K.get_value(weight))))
              for weight in prunable_weights
          ])))
  print("\n")


def get_model_sparsity(model, per_layer=False, allow_list=None):
  """Calculates the sparsity of the model's weights and biases.

  Quantizes the model weights using model_save_quantized_weights (but does not
    save the quantized weights) before calculating the proportion of weights and
    biases set to zero.

  Arguments:
      model: The model to use to calculate sparsity. Assumes that this is a
          QKeras model with trained weights.
      per_layer: If to return a per-layer breakdown of sparsity
      allow_list: A list of layer class names that sparsity will be calculated
        for. If set to None, a default list will be used.

  Returns:
      A float value representing the proportion of weights and biases set to
      zero in the quantized model. If per_layer is True, it also returns a
      per-layer breakdown of model sparsity formatted as a list of tuples in the
      form (<layer name>, <sparsity proportion>)
  """
  # Checks if to use a default list of allowed layers to calculate sparsity.
  if allow_list is None:
    allow_list = [
        "QDense", "Dense", "QConv1D", "Conv1D", "QConv2D", "Conv2D",
        "QDepthwiseConv2D", "DepthwiseConv2D",
        "QSeparableConv1D", "SeparableConv1D",
        "QSeparableConv2D", "SeparableConv2D", "QOctaveConv2D",
        "QSimpleRNN", "RNN", "QLSTM", "QGRU",
        "QConv2DTranspose", "Conv2DTranspose",
        "QConv2DBatchnorm", "QDepthwiseConv2DBatchnorm",
    ]

  # Quantizes the model weights for a more accurate sparsity calculation.
  model_save_quantized_weights(model)

  # Calculates the sparsity layer by layer.
  layer_sparsity = []
  total_sparsity = 0.
  all_weights = []
  for layer in model.layers:
    if hasattr(layer, "quantizers") and layer.__class__.__name__ in allow_list:
      if layer.__class__.__name__ in [
          "QConv2DBatchnorm", "QDepthwiseConv2DBatchnorm"]:
        weights_to_examine = layer.get_folded_weights()
      else:
        weights_to_examine = layer.get_weights()

      layer_weights = []
      for weight in weights_to_examine:
        try:
          weight_numpy = weight.ravel()
        except AttributeError:
          # In case of EagerTensor.
          weight_numpy = weight.numpy().ravel()
        layer_weights.append(weight_numpy)
        all_weights.append(weight_numpy)
      layer_weights = np.concatenate(layer_weights)
      layer_sparsity.append((layer.name, np.mean(layer_weights == 0)))

  if len(all_weights) > 0:
    # Average the sparsity for the entire model.
    all_weights = np.concatenate(all_weights)
    total_sparsity = np.mean(all_weights == 0)
  if per_layer:
    return (total_sparsity, layer_sparsity)
  else:
    return total_sparsity


def quantized_model_debug(model, X_test, plot=False, plt_instance=None):
  """Debugs and plots model weights and activations.

  Args:
    model: The QKeras model to debug
    X_test: The sample data to use to give to model.predict
    plot: Bool. If to plot the results.
    plt_instance: A matplotlib.pyplot instance used to plot in an IPython
      environment.
  """
  assert (plt_instance and plot) or not plot, (
      "plt_instance is required if plt is True")

  outputs = []
  output_names = []

  for layer in model.layers:
    if layer.__class__.__name__ in REGISTERED_LAYERS:
      output_names.append(layer.name)
      outputs.append(layer.output)

  model_debug = Model(inputs=model.inputs, outputs=outputs)

  y_pred = model_debug.predict(X_test)

  print("{:30} {: 8.4f} {: 8.4f}".format(
      "input", np.min(X_test), np.max(X_test)))

  for n, p in zip(output_names, y_pred):
    layer = model.get_layer(n)
    if (layer.__class__.__name__ in "QActivation" or
        layer.__class__.__name__ in "QAdaptiveActivation"):
      alpha = get_weight_scale(layer.activation, p)
    else:
      alpha = 1.0
    print(
        "{:30} {: 8.4f} {: 8.4f}".format(n, np.min(p / alpha),
                                         np.max(p / alpha)),
        end="")
    if alpha != 1.0:
      print(" a[{: 8.4f} {:8.4f}]".format(np.min(alpha), np.max(alpha)))
    if plot and layer.__class__.__name__ in [
        "QConv1D", "QConv2D", "QConv2DTranspose", "QDense", "QActivation",
        "QAdaptiveActivation", "QSimpleRNN", "QLSTM", "QGRU", "QBidirectional",
        "QSeparableConv1D", "QSeparableConv2D"
    ]:
      plt_instance.hist(p.flatten(), bins=25)
      plt_instance.title(layer.name + "(output)")
      plt_instance.show()
    alpha = None

    if layer.__class__.__name__ not in [
        "QConv2DBatchnorm", "QDepthwiseConv2DBatchnorm"]:
      weights_to_examine = layer.get_weights()
    else:
      weights_to_examine = layer.get_folded_weights()

    for i, weights in enumerate(weights_to_examine):
      if hasattr(layer, "get_quantizers") and layer.get_quantizers()[i]:
        weights = K.eval(layer.get_quantizers()[i](K.constant(weights)))
        if i == 0 and layer.__class__.__name__ in [
            "QConv1D", "QConv2D", "QConv2DTranspose", "QDense",
            "QSimpleRNN", "QLSTM", "QGRU",
            "QSeparableConv1D", "QSeparableConv2D",
            "QConv2DBatchnorm", "QDepthwiseConv2DBatchnorm"
        ]:
          alpha = get_weight_scale(layer.get_quantizers()[i], weights)
          # if alpha is 0, let's remove all weights.
          alpha_mask = (alpha == 0.0)
          weights = np.where(alpha_mask, weights * alpha, weights / alpha)
          if plot:
            plt_instance.hist(weights.flatten(), bins=25)
            plt_instance.title(layer.name + "(weights)")
            plt_instance.show()
      print(" ({: 8.4f} {: 8.4f})".format(np.min(weights), np.max(weights)),
            end="")
    if alpha is not None and isinstance(alpha, np.ndarray):
      print(" a({: 10.6f} {: 10.6f})".format(
          np.min(alpha), np.max(alpha)), end="")
    print("")


def quantized_model_dump(model,
                         x_test,
                         output_dir=None,
                         layers_to_dump=[]):
  """Dumps tensors of target layers to binary files.

  Arguments:
    model: QKeras model object.
    x_test: numpy type, test tensors to generate output tensors.
    output_dir: a string for the directory to hold binary data.
    layers_to_dump: a list of string, specified layers by layer
      customized name.
  """
  outputs = []
  y_names = []

  if not output_dir:
    with tempfile.TemporaryDirectory() as output_dir:
      print("temp dir", output_dir)

  if not os.path.exists(output_dir):
    os.makedirs(output_dir)
    print("create dir", output_dir)

  for layer in model.layers:
    if not layers_to_dump or layer.name in layers_to_dump:
      y_names.append(layer.name)
      outputs.append(layer.output)

  # Gather the tensor outputs from specified layers at layers_to_dump.
  model_debug = Model(inputs=model.inputs, outputs=outputs)
  y_pred = model_debug.predict(x_test)

  # Dumps tensors to files.
  for name, tensor_data in zip(y_names, y_pred):
    filename = os.path.join(output_dir, name + ".bin")
    print("writing the layer output tensor to ", filename)
    with open(filename, "w") as fid:
      tensor_data.astype(np.float32).tofile(fid)



# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Export quantizer package."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .accumulator_factory import AccumulatorFactory
from .multiplier_factory import MultiplierFactory
from .multiplier_impl import IMultiplier, FloatingPointMultiplier, FixedPointMultiplier, Mux, AndGate, Adder, XorGate, Shifter
from .accumulator_impl import IAccumulator, FloatingPointAccumulator, FixedPointAccumulator
from .quantizer_impl import IQuantizer, QuantizedBits, Binary, QuantizedRelu, Ternary, FloatingPoint, PowerOfTwo, ReluPowerOfTwo
from .quantizer_factory import QuantizerFactory
from .qbn_factory import QBNFactory
from .merge_factory import MergeFactory
from .divider_factory import IDivider
from .subtractor_factory import ISubtractor



# Lint as: python3
# ==============================================================================
# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements support for auto-quantization."""

import collections
import json
import os
import re
import copy
from absl import logging
import keras_tuner as kt
from keras_tuner import HyperModel
from keras_tuner import BayesianOptimization
from keras_tuner import Hyperband
from keras_tuner import RandomSearch
import numpy as np
import six
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.metrics import binary_accuracy
from tensorflow.keras.metrics import categorical_accuracy
from tensorflow.keras.metrics import sparse_categorical_accuracy
from qkeras.autoqkeras.forgiving_metrics import forgiving_factor  # pylint: disable=line-too-long
from qkeras.autoqkeras.forgiving_metrics import ForgivingFactor  # pylint: disable=line-too-long
from qkeras.autoqkeras.quantization_config import default_quantization_config  # pylint: disable=line-too-long
from qkeras.autoqkeras.utils import print_qmodel_summary
from qkeras.utils import clone_model
from qkeras.utils import model_quantize


# AutoQKHyperModel is implemented on top of keras_tuner
# It basically creates a quantized model based on some rules
# and it computes a acc_delta that boosts the accuracy when
# choosing smaller models.

# Boosting function behaves like this.
# We use the following formula to compute the decrease factor:
#   reference_size: number of parameters + activations of the model,
#     assuming an 8-bit implementation.
#   trial_size: number of parameters + activations of trial.
#
#   1) First, we compute how many times we decresed/increased the model
#     i = log(reference_size / trial_size) / log(rate)
#
#   2) Then, we use delta_p / delta_n if model is smaller/bigger
#      than reference model.
#
#      delta = i * (
#          (i < 0) * delta_n + (i >= 0) * delta_p
#      )
#
#   3) the accuracy of the model (score) is adjusted by acc * delta
#
#   The delta "boosts" the accuracy to allow worse model to be
#   chosen by hypermodel tuner.
#

REGISTERED_LAYERS = ["Dense", "Conv1D", "Conv2D", "DepthwiseConv2D",
                     "SimpleRNN", "LSTM", "GRU", "Bidirectional",
                     "Conv2DTranspose", "SeparableConv1D", "SeparableConv2D"]

Q_LAYERS = list(map(lambda x : 'Q' + x, REGISTERED_LAYERS))

SEQUENCE_LAYERS = ["SimpleRNN", "LSTM", "GRU", "Bidirectional"]

class AutoQKHyperModel(HyperModel):
  """Creates an hypermodel to attempt to quantize a reference model.

     Arguments:
       model: Model to be quantized.
       metrics: List of metrics to be used.
       custom_objects: Custom objects used by Keras during quantization.
       target: Secondary metric to chase during search ("bits" or "energy").
       transfer_weights: if true, transfer weights from unquantized model.
       frozen_layers: if true, these layers will not be quantized but
         weights transferred from original model.
       activation_bits: parameter to be used by 'model_quantize'.
       limit: limit the number of bits in quantizers, specified as dictionary.
       tune_filters: one of "block", "layer", "none" for tuning entire
         network, each layer separately, or no tuning.
       tune_filters_exceptions: name of layers that will not be tuned.
       layer_indexes: we only quantize layers whose ids are in layer_indexes.
       learning_rate_optimizer: if true, we optimize learning rate along with
         other parameters.
       head_name: specify which head to calcuate score/trial-size from in
         autoqkeras
       quantization_config: dictionary containing configuration of
         quantizers for kernel, bias and activation.
       extend_model_metrics: If to append the trial size and score metrics to
         model metrics, which are used for AutoQKeras to determine the quality
         of a model.

     Returns:
       quantized model in trial and boosted accuracy function compiled
       into quantized model.
  """

  def __init__(
      self, model, metrics, custom_objects=None, target=None,
      transfer_weights=False, frozen_layers=None, activation_bits=4, limit=None,
      tune_filters="none", tune_filters_exceptions=None,
      layer_indexes=None, learning_rate_optimizer=False,
      head_name=None, quantization_config=None, extend_model_metrics=True,
  ):
    self.model = model
    self.metrics = metrics
    self.custom_objects = custom_objects if custom_objects else {}

    self.target = target

    self.reference_size = self.target.get_reference(model)

    self.transfer_weights = transfer_weights
    self.frozen_layers = frozen_layers if frozen_layers else []
    self.activation_bits = activation_bits
    self.head_name = head_name
    self.extend_model_metrics = extend_model_metrics
    # make sure we have at least 3 elements in list
    # first one for kernel, second one for bias and thid one for activations.
    #
    # limit is in the format, where default replaces missing values:
    # '{
    #      "Conv2D":[weight,bias,activation],
    #      "RNN":[weight,bias,recurrent,activation],
    #      "Dense":[weight,bias,activation],
    #      "Activation":[activation]
    #      "default": value
    #  }'

    if limit is None:
      self.limit = {}
    else:
      self.limit = limit

    self.groups = {}

    assert isinstance(self.limit, dict)

    if self.limit.get("default", None) is None:
      default = 8
    else:
      default = self.limit["default"]

    # make sure we have entries for every type of layer we process
    self._adjust_limit(default)

    print("Limit configuration:" + json.dumps(self.limit))

    assert tune_filters in ["block", "layer", "none"]

    self.tune_filters = tune_filters
    self.tune_filters_exceptions = re.compile(tune_filters_exceptions)

    self.layer_indexes = layer_indexes
    self.learning_rate_optimizer = learning_rate_optimizer

    # load quantizer types for each type of quantizer
    if quantization_config is None:
      self.quantization_config = default_quantization_config
    else:
      self.quantization_config = quantization_config

  def _adjust_limit(self, default):
    """Makes sure limit has all the fields required."""
    if isinstance(default, list):
      assert 3 <= len(default) <= 4 
    else:
      default = [default] * 3

    # we consider that if name is not there, we will ignore the layer
    for name in REGISTERED_LAYERS:
      if name in self.limit:
        length = len(self.limit[name])
        if length < 4 and name in SEQUENCE_LAYERS:
          assert len(default) == 4
          self.limit[name] = self.limit[name] + default[length:]
        elif length < 3:
          # No recurrent limit needed for non recurrent layers
          self.limit[name] = self.limit[name] + default[length:2] + default[-1:]

  def _n(self, name, s_list):
    """Creates a unique name for the tuner."""
    return name + "_".join([str(v) for v in s_list])

  def _get_quantizer(self, hp, head, layer_name, layer_class_name,
                     i_list=None, is_kernel=True, is_linear=False):
    """Gets a quantizer randomly for kernels/bias/activations."""

    # first pick up which group we belong to.

    if not i_list:
      i_list = []

    if is_linear:
      # linear quantizers
      field_name = "linear"
      kq = self.quantization_config["linear"]
      index = 0
      q_list = list(kq.keys())
      q_dict = kq
    elif "kernel" in head:
      # kernel quantizers
      field_name = "kernel"
      kq = self.quantization_config["kernel"]
      index = 0
      q_list = list(kq.keys())
      q_dict = kq
    elif "bias" in head:
      # bias quantizers
      field_name = "bias"
      bq = self.quantization_config["bias"]
      index = 1
      q_list = list(bq.keys())
      q_dict = bq
    elif "pointwise_kernel" in head: # limit is same as kernel
      # pointwise kernel quantizers
      field_name = "pointwise_kernel"
      kq = self.quantization_config["pointwise_kernel"]
      index = 2
      q_list = list(kq.keys())
      q_dict = kq
    elif "recurrent_kernel" in head: # limit is same as kernel
      # recurrent kernel quantizers
      field_name = "recurrent_kernel"
      kq = self.quantization_config["recurrent_kernel"]
      index = 2
      q_list = list(kq.keys())
      q_dict = kq
    elif "recurrent_activation" in head: # limit is same as kernel
      # recurrent activation quantizers
      field_name = "recurrent_activation"
      raq = self.quantization_config["recurrent_activation"]
      index = -1
      q_list = list(raq.keys())
      q_dict = raq
    else:
      # activation quantizers
      field_name = "activation"
      aq = self.quantization_config["activation"]
      index = -1
      q_list = list(aq.keys())
      q_dict = aq

    # we first we search for layer name. If it is not there, we switch to
    # layer class name.

    found_pattern = False
    name = layer_class_name
    count = -1
    for i, pattern in enumerate(self.limit):
      if re.match(pattern, layer_name):
        found_pattern = True
        name = pattern
        count = i
        break

    # for partially quantized networks we may not have
    # the layer class name in the set.

    if name == layer_class_name and name not in self.limit:
      return None, -1

    # groups is a dictionary that contains dictionary of the
    # patterns so that we can group everything together

    if found_pattern:
      if name in self.groups and index in self.groups[name]:
        return self.groups[name][index]

      # not there, let's use a different name for
      # the head and field
      head = "qk_group_" + str(count) + "_" + field_name
      head = name + "_" + field_name

    # limit group can be a list of quantizers or a
    # number that tells us maximum number of bits

    if isinstance(self.limit[name][index], list):
      # we assume this is a subset of the q_keys
      # entry in quantization_config will be like:
      #   "Conv2D": [ ["q1", "q2", "q3"], ... ]
      #
      # we always assume this list is a subset of
      # the original list or we will raise an
      # error.

      q_list = self.limit[name][index]
      q_dict = {
          key: q_dict[key] for key in q_list
      }
    else:
      q_dict = {
          key: value for (key, value) in q_dict.items()
          if value <= self.limit[name][index]
      }
      q_list = list(q_dict.keys())

    # didn't found a match in groups, create one.

    if len(q_list) == 1:
      q_name = hp.Fixed(self._n(head + "_quantizer", i_list), q_list[0])
    else:
      q_name = hp.Choice(self._n(head + "_quantizer", i_list), q_list)

    if found_pattern:
      if name not in self.groups:
        self.groups[name] = {index: (q_name, q_dict[q_name])}
      else:
        self.groups[name][index] = (q_name, q_dict[q_name])

    return (q_name, q_dict[q_name])

  def quantize_model(self, hp):
    """Quantize model by hyperparameter search and extracting size schema."""

    # configuration for quantization.
    q_dict = {}

    model = clone_model(self.model, self.custom_objects)

    fanin = []

    filter_range = [0.5, 0.75, 1.0, 1.5, 2.0]

    # network_filters=hp.Choice(...) should only be defined if we are sure
    # current blocks has any layer that need filter sweep.
    # Otherwise, when no layer needs filter sweep and a hp variable is defined,
    # there will be uneffective trials that loop around the network
    # filter range, even though none of the filter sweep was ever applied to
    # any layers. Therfore, we use filter_sweep_enabled to mark if any layer
    # in current block needs filter sweep.
    kernel_quantizer_dict = {}
    filter_sweep_enabled = False
    for layer in model.layers:
      if layer.__class__.__name__ in REGISTERED_LAYERS:
        kernel_quantizer, bits = self._get_quantizer(
            hp, layer.name + "_kernel", layer.name, layer.__class__.__name__,
            is_kernel=True)

        kernel_quantizer_dict[layer.name] = (kernel_quantizer, bits)

        # kernel_quantizer is not None ->  layer in the current block need
        # to be quantized
        if kernel_quantizer:
          if (
              not filter_sweep_enabled and self.tune_filters in
              ["layer", "block"]
              and not self.tune_filters_exceptions.search(layer.name) and
              layer.__class__.__name__ in
              ["Dense", "Conv1D", "Conv2D", "Conv2DTranspose"]
          ):
            filter_sweep_enabled = True

        if layer.__class__.__name__ in SEQUENCE_LAYERS:
          recurrent_quantizer, _ = self._get_quantizer(
            hp, layer.name + "_recurrent_kernel", layer.name, layer.__class__.__name__,
            is_kernel=True)

        if layer.__class__.__name__ in ["SeparableConv1D", "SeparableConv2D"]:
          pointwise_quantizer, _ = self._get_quantizer(
            hp, layer.name + "_pointwise_kernel", layer.name, layer.__class__.__name__,
            is_kernel=True)

    if self.tune_filters == "block" and filter_sweep_enabled:
      network_filters = hp.Choice(
          "network_filters",
          values=filter_range,
          default=1.0
      )
    else:
      network_filters = 1.0

    for layer_id, layer in enumerate(model.layers):

      # we can use these indexes to disable some layers, like the last
      # layer

      if self.layer_indexes is not None and layer_id not in self.layer_indexes:
        continue

      layer_d = {}

      if layer.__class__.__name__ in Q_LAYERS:
        weights = layer.get_weights()[0]
        if (
            layer.get_quantizers()[0] and
            hasattr(layer.get_quantizers()[0], "bits")
        ):
          bits = layer.get_quantizers()[0].bits
        else:
          bits = 8
        fanin.append(np.prod(weights.shape[:-1]) * (8. - bits) / 8.)
        
      if layer.__class__.__name__ in REGISTERED_LAYERS:
        # difference between depthwise and the rest is just the name
        # of the kernel.
        if layer.__class__.__name__ in [
            "DepthwiseConv2D", "SeparableConv1D", "SeparableConv2D"
        ]:
          kernel_name = "depthwise_quantizer"
        else:
          kernel_name = "kernel_quantizer"

        # sample kernel quantizer.
        (kernel_quantizer, bits) = kernel_quantizer_dict[layer.name]

        if not kernel_quantizer:
          continue

        # process fanin here

        if bits < 8:
          weights = layer.get_weights()[0]
          fanin.append(np.prod(weights.shape[:-1]) * (8. - bits) / 8.)

        # we only want to do that if we are going to quantize layer
        if (
            self.tune_filters in ["layer", "block"] and
            not self.tune_filters_exceptions.search(layer.name) and
            layer.__class__.__name__ in [
                "Dense", "Conv1D", "Conv2D", "Conv2DTranspose",
                "SeparableConv1D", "SeparableConv2D"
            ]
        ):
          if self.tune_filters == "layer":
            layer_filters = hp.Choice(
                "network_filters_" + layer.name,
                values=filter_range,
                default=1.0
            )
          else:
            layer_filters = network_filters

          if layer.__class__.__name__ == "Dense":
            layer.units = max(int(layer.units * layer_filters), 1)
          elif layer.__class__.__name__ in [
              "Conv1D", "Conv2D", "Conv2DTranspose",
              "SeparableConv1D", "SeparableConv2D"
          ]:
            layer.filters = max(int(layer.filters * layer_filters), 1)

        layer_d[kernel_name] = kernel_quantizer

        if layer.__class__.__name__ in SEQUENCE_LAYERS:
          layer_d['recurrent_quantizer'] = recurrent_quantizer

        if layer.__class__.__name__ in ["SeparableConv1D", "SeparableConv2D"]:
          layer_d['pointwise_quantizer'] = pointwise_quantizer

        if layer.__class__.__name__ in ["LSTM", "GRU", "Bidirectional"]:
          layer_d['recurrent_activation'], _  = self._get_quantizer(
              hp, layer.name + "_recurrent_activation", layer.name,
              layer.__class__.__name__, is_kernel=False)

        # if we use bias, sample quantizer.
        if layer.__class__.__name__ == "Bidirectional":
          layer_d["bias_quantizer"], bits = self._get_quantizer(
              hp, layer.name + "_bias", layer.name, layer.__class__.__name__,
              is_kernel=False)
          layer_d["activation"], bits = self._get_quantizer(
              hp, layer.name + "_activation", layer.name,
              layer.__class__.__name__, is_kernel=False)
          q_dict[layer.name] = layer_d 
        else:
          if layer.use_bias:
            layer_d["bias_quantizer"], bits = self._get_quantizer(
                hp, layer.name + "_bias", layer.name, layer.__class__.__name__,
                is_kernel=False)

          # if activation is not linear/softmax we need to process it.
          if layer.activation is None:
            is_softmax = False
            is_linear = False
          else:
            if isinstance(layer.activation, six.string_types):
              is_softmax = layer.activation == "softmax"
              is_linear = layer.activation == "linear"
            else:
              is_softmax = layer.activation.__name__ == "softmax"
              is_linear = layer.activation.__name__ == "linear"

          if not is_softmax and not is_linear:
            layer_d["activation"], bits = self._get_quantizer(
                hp, layer.name + "_activation", layer.name,
                layer.__class__.__name__, is_kernel=False)

          q_dict[layer.name] = layer_d

      elif layer.__class__.__name__ in ["Reshape"]:
        # we cannot handle fine tuning filters per layer right now.
        assert self.tune_filters in ["none", "block"]

        # we need to make sure this pattern exists, this should only occur for
        # "scheduler", so the name will be complete and not a pattern.

        if (
            self.tune_filters == "none" or
            layer.name not in self.limit or
            self.tune_filters_exceptions.search(layer.name)
        ):
          continue

        if K.image_data_format() == "channels_last":
          layer.target_shape = layer.target_shape[:-1] + (
              min(int(layer.target_shape[-1] * network_filters), 1),)
        else:
          layer.target_shape = (int(layer.target_shape[0] * network_filters),
                                ) + layer.target_shape[1:]

      elif layer.__class__.__name__ in ["Activation"]:
        if isinstance(layer.activation, six.string_types):
          is_linear = layer.activation == "linear"
          is_softmax = layer.activation == "softmax"
        else:
          is_linear = layer.activation.__name__ == "linear"
          is_softmax = layer.activation.__name__ == "softmax"

        # if it is a linear activation, we will notify the
        # quantizer we are searching for linear type of
        # quantizers

        if not is_softmax:
          activation, bits = self._get_quantizer(
              hp, layer.name + "_activation", layer.name,
              layer.__class__.__name__, is_kernel=False,
              is_linear=is_linear)

          if not activation:
            continue

          # look at documentation on model_quantize
          q_dict[layer.name] = activation
      elif layer.__class__.__name__ in self.limit:
        # mark it for conversion
        q_dict[layer.name] = {}
      else:
        for pattern in self.limit:
          if re.match(pattern, layer.name):
            q_dict[layer.name] = {}
            break

    q_model = model_quantize(
        model, q_dict, self.activation_bits,
        custom_objects=self.custom_objects,
        transfer_weights=self.transfer_weights)

    return q_model, fanin

  def build(self, hp):
    """Builds hyperparameterized quantized model."""

    self.groups = {}

    # we are not using the fanin right now.

    q_model, _ = self.quantize_model(hp)

    # transfer weights from previous run as we know we will not
    if self.learning_rate_optimizer:
      # if learning_rate_optimizer, we try to transfer weights from previous run
      print("... freezing layers {}.".format(", ".join(self.frozen_layers)))
      for layer_name in self.frozen_layers:
        o_weights = self.model.get_layer(layer_name).get_weights()
        layer = q_model.get_layer(layer_name)
        # don't know if setting trainable to False is good or not yet
        # try to do "soft-freeze" by transferring weights. More experiments
        # needed before we decide what to do.
        # layer.trainable = False
        weights = layer.get_weights()
        # because we can be changing number of layers, we do not know
        # if we can really use some of the weights or not.
        equal_layer = True
        for w in range(len(o_weights)):
          if o_weights[w].shape != weights[w].shape:
            equal_layer = False
            break
        if equal_layer:
          layer.set_weights(o_weights)

    self.trial_size = self.target.get_trial(q_model)

    # we will use a boosted accuracy computation

    delta = self.target.delta()

    # by default, we use the first metric specified by the
    # user to be the target metric.
    if not self.metrics:
      score_metric = None
    elif isinstance(self.metrics, dict):
      if not self.head_name:
      # if head_name not provided, find the first metric from the dict
        score_key = list(self.metrics.keys())[0]
      else:
        # find the metric assoicated with the head_name
        score_key = self.head_name
      score_metric = self.metrics[score_key]
      if isinstance(score_metric, list):
        score_metric = score_metric[0]
    elif isinstance(self.metrics, list):
      score_metric = self.metrics[0]

    self.score = AutoQKHyperModel.adjusted_score(
        self, delta, score_metric)

    # some papers suggest that we use learning_rate * sqrt(fanin) / layer
    # we cannot do that right now, but we can definitely do that
    # if we are quantizing one layer at a time
    #
    # https://arxiv.org/pdf/1511.00363.pdf

    # we use the magic number to smooth out the average
    total_factor = self.target.get_total_factor()
    delta_lr = 1.0 + (total_factor < 0) * total_factor

    # we assume model has been compiled at least.

    lr = float(self.model.optimizer.lr.numpy())

    # we assume that delta_lr can lower lr to accommodate
    # for more quantization
    #
    # if learning rate scheduler is used, we assume the callback to manage
    # learning rate. Just set it to constant.

    if self.learning_rate_optimizer:
      lr_range = list(lr * np.linspace(delta_lr, 1.1, 5))
      lr_choice = hp.Choice("learning_rate", lr_range)
      self.model.optimizer.learning_rate = lr_choice
    else:
      lr_choice = lr
      print("learning_rate: {}".format(lr))

    optimizer = self.model.optimizer

    q_model.summary()

    metrics = self.metrics

    # extend metrics by including score and trial_size metrics
    if self.extend_model_metrics:
      ext_metrics = copy.deepcopy(metrics)
      if isinstance(ext_metrics, dict):
        # for dict, add trial_size_metric and score metric to target output
        if not self.head_name:
          # if head_name not provided, find the first metric from the dict
          score_key = list(ext_metrics.keys())[0]
        else:
          # find the metric assoicated with the head_name
          score_key = self.head_name
        score_metric = ext_metrics[score_key]
        if isinstance(score_metric, list):
          score_metric += [self.trial_size_metric(self.trial_size), self.score]
        else:
          score_metric = [score_metric]
          score_metric += [self.trial_size_metric(self.trial_size), self.score]
        ext_metrics[score_key] = score_metric
      else:
        ext_metrics += [
            self.trial_size_metric(self.trial_size),
            self.score]
      metrics = ext_metrics

    q_model.compile(
        optimizer=optimizer,
        loss=self.model.loss,
        metrics=metrics
    )
    self.q_model = q_model

    # this just prints a summary of the quantization for debugging
    # purposes

    self.target.print_stats()
    print_qmodel_summary(q_model)

    return q_model

  @staticmethod
  def adjusted_score(hyper_model, delta, metric_function=None):
    def score(y_true, y_pred):
      y_t_rank = len(y_true.shape.as_list())
      y_p_rank = len(y_pred.shape.as_list())
      y_t_last_dim = y_true.shape.as_list()[-1]
      y_p_last_dim = y_pred.shape.as_list()[-1]

      is_binary = y_p_last_dim == 1
      is_sparse_categorical = (
          y_t_rank < y_p_rank or y_t_last_dim == 1 and y_p_last_dim > 1)

      if isinstance(metric_function, six.string_types):
        if metric_function in ["accuracy", "acc"]:
          if is_binary:
            metric = binary_accuracy(y_true, y_pred)
          elif is_sparse_categorical:
            metric = sparse_categorical_accuracy(y_true, y_pred)
          else:
            metric = categorical_accuracy(y_true, y_pred)
        else:
          metric = categorical_accuracy(y_true, y_pred)
      else:
        metric = metric_function(y_true, y_pred)

      return K.cast(metric * (1.0 + delta), K.floatx())

    if not metric_function:
      metric_function = "accuracy"

    return score

  @staticmethod
  def trial_size_metric(trial_size):
    def trial(y_true, y_pred):  # pylint: disable=unused-argument
      return K.cast(trial_size, K.floatx())
    return trial


class AutoQKeras:
  """Performs autoquantization in Keras model.

     Arguments:
       model: Model to be quantized.
       metrics: List of metrics to be used.
       custom_objects: Custom objects used by Keras during quantization.
       goal: Metric to compute secondary goal of search (bits or energy)
       output_dir: name of output directory to store results.
       mode: random, hyperband or bayesian used by keras_tuner.
       custom_tuner: The Keras Tuner class to use to search hyperparams
       transfer_weights: if true, transfer weights from unquantized model.
       frozen_layers: if true, these layers will not be quantized but
         weights transferred from original model.
       activation_bits: parameter to be used by 'model_quantize'.
       limit: limit the number of bits in quantizers specified as a dictionary.
       tune_filters: one of "block", "layer", "none" for tuning entire
         network, each layer separately, or no tuning.
       tune_filters_exceptions: name of layers that will not be tuned.
       layer_indexes: indexes of layers we will quantize.
       learning_rate_optimizer: if true, user will provide lr scheduler
         callback.
       quantization_config: file name of dictionary containing configuration of
         quantizers for kernel, bias and activation.
       head_name: specify which head to calcuate score/trial-size from in
         autoqkeras
       score_metric: Str. Optional metric name to use to evaluate the trials.
         Defaults to val_score
       tuner_kwargs: parameters for keras_tuner depending on whether
         mode is random, hyperband or baeysian. Please refer to the
         documentation of kerstuner Tuners.
  """

  def __init__(
      self, model, metrics=None, custom_objects=None, goal=None,
      output_dir="result", mode="random", custom_tuner=None,
      transfer_weights=False, frozen_layers=None, activation_bits=4,
      limit=None, tune_filters="none",
      tune_filters_exceptions=None, learning_rate_optimizer=False,
      layer_indexes=None, quantization_config=None, overwrite=True,
      head_name=None, score_metric=None, **tuner_kwargs):

    # Collect input arguments to AutoQKeras for usage by custom tuner
    autoqkeras_input_args = locals()

    if not metrics:
      metrics = []

    if not custom_objects:
      custom_objects = {}

    # goal: { "type": ["bits", "energy"], "params": {...} } or ForgivingFactor
    #   type
    # For type == "bits":
    #   delta_p: increment (in %) of the accuracy if trial is smaller.
    #   delta_n: decrement (in %) of the accuracy if trial is bigger.
    #   rate: rate of decrease/increase in model size in terms of bits.
    #   input_bits; size of input tensors.
    #   output_bits; size of output tensors.
    #   stress: parameter to reduce reference size to force tuner to
    #     choose smaller models.
    #   config: configuration on what to compute for each layer
    #     minimum configuration is { "default": ["parameters", "activations"] }

    # use simplest one - number of bits
    if not goal:
      goal = {
          "type": "bits",
          "params": {
              "delta_p": 8.0,
              "delta_n": 8.0,
              "rate": 2.0,
              "stress": 1.0,
              "input_bits": 8,
              "output_bits": 8,
              "ref_bits": 8,
              "config": {
                  "default": ["parameters", "activations"]
              }
          }
      }

    self.overwrite = overwrite

    # for multi-head model, we need to specify which head(/output) that
    # score and trial metric needs to calculate from
    self.head_name = head_name

    # if we have not created it already, create new one.
    if not isinstance(goal, ForgivingFactor):
      target = forgiving_factor[goal["type"]](**goal["params"])
    else:
      target = goal

    # if no metrics were specified, we want to make sure we monitor at least
    # accuracy.
    if not metrics:
      metrics = ["acc"]

    self.hypermodel = AutoQKHyperModel(
        model, metrics, custom_objects, target,
        transfer_weights=transfer_weights,
        frozen_layers=frozen_layers,
        activation_bits=activation_bits,
        limit=limit,
        tune_filters=tune_filters,
        tune_filters_exceptions=tune_filters_exceptions,
        layer_indexes=layer_indexes,
        learning_rate_optimizer=learning_rate_optimizer,
        head_name=head_name,
        quantization_config=quantization_config
    )

    # right now we create unique results directory
    idx = 0
    name = output_dir
    if self.overwrite:
      while os.path.exists(name):
        idx += 1
        name = output_dir + "_" + str(idx)
    output_dir = name
    self.output_dir = output_dir

    if score_metric is None:
      if self.head_name:
        score_metric = "val_" + self.head_name + "_score"
      else:
        score_metric = "val_score"
    assert mode in ["random", "bayesian", "hyperband"]
    if custom_tuner is not None:
      self.tuner = custom_tuner(
          self.hypermodel,
          autoqkeras_config=autoqkeras_input_args,
          objective=kt.Objective(score_metric, "max"),
          project_name=output_dir,
          **tuner_kwargs)
    elif mode == "random":
      self.tuner = RandomSearch(
          self.hypermodel,
          objective=kt.Objective(score_metric, "max"),
          project_name=output_dir,
          **tuner_kwargs)
    elif mode == "bayesian":
      self.tuner = BayesianOptimization(
          self.hypermodel,
          objective=kt.Objective(score_metric, "max"),
          project_name=output_dir,
          **tuner_kwargs)
    elif mode == "hyperband":
      self.tuner = Hyperband(
          self.hypermodel,
          objective=kt.Objective(score_metric, "max"),
          project_name=output_dir,
          **tuner_kwargs)
    else:
      pass

    self.tuner.search_space_summary()

  def _has_earlystopping(self, callbacks):
    """Check if EarlyStopping has been defined or not."""
    if callbacks is None:
      return False

    for callback in callbacks:
      if isinstance(callback, tf.keras.callbacks.EarlyStopping):
        return True
    return False

  def history(self, number_of_trials=-1):
    """Returns the history of the model search."""
    trials = self.tuner.oracle.get_best_trials(number_of_trials)
    state = [trial.get_state() for trial in trials]

    result = {}
    result["score"] = [
        state[i]["score"] for i in range(len(state))
        if trials[i].score is not None
    ]
    for i in range(len(state)):
      if trials[i].score is not None:
        keys = state[i]["metrics"]["metrics"].keys()

        for key in keys:
          if key != "score" and not key.startswith(
              "val_") and key != "loss" and key != "trial":

            cur_accuracy = state[i]["metrics"]["metrics"][key][
                "observations"][0]["value"][0]
            if "val_" + key in state[i]["metrics"]["metrics"].keys():
              cur_val_accuracy = state[i]["metrics"]["metrics"]["val_" + key][
                  "observations"][0]["value"][0]
            else:
              cur_val_accuracy = None

            # only update result if both key and val_key exist
            if cur_val_accuracy:
              if key not in result.keys():
                result[key] = [cur_accuracy]
                result["val_" + key] = [cur_val_accuracy]
              else:
                result[key].append(cur_accuracy)
                result["val_" + key].append(cur_val_accuracy)

    if self.head_name:
      trial_from_output = self.head_name + "_trial"
    else:
      trial_from_output = "trial"
    result["trial_size"] = [
        state[i]["metrics"]["metrics"][trial_from_output]["observations"][0]
        ["value"][0] for i in range(len(state)) if trials[i].score is not None
    ]

    return result

  def fit(self, *fit_args, **fit_kwargs):
    """Invokes tuner fit algorithm."""

    callbacks = fit_kwargs.get("callbacks", None)

    if callbacks is None:
      callbacks = []

    epochs = fit_kwargs.get("epochs", None)

    if epochs is None:
      epochs = 10

    if not self._has_earlystopping(callbacks):
      callbacks = callbacks + [
          tf.keras.callbacks.EarlyStopping(
              "val_loss", patience=min(20, epochs//5))
      ]
      fit_kwargs["callbacks"] = callbacks

    self.tuner.search(*fit_args, **fit_kwargs)

  @staticmethod
  def get_best_lr(qmodel):
    """Extracts best lr of model."""
    return qmodel.optimizer.lr.numpy()

  def get_best_model(self):
    params = self.tuner.get_best_hyperparameters()[0]

    q_model = self.tuner.hypermodel.build(params)

    self.learning_rate = q_model.optimizer.lr.numpy()

    return q_model

  def get_learning_rate(self):
    return self.learning_rate


class AutoQKerasScheduler:
  """Performs autoquantization one layer/group at a time.

     Arguments:
       model: Model to be quantized.
       metrics: List of metrics to be monitored.
       custom_objects: Custom objects used by Keras during quantization.
       goal: Metric to compute secondary goal of search (bits or energy)
       output_dir: name of output directory to store results.
       mode: random, hyperband or bayesian used by keras_tuner.
       transfer_weights: if true, transfer weights from unquantized model.
       activation_bits: parameter to be used by 'model_quantize'.
       limit: limit the number of bits in quantizers specified as a dictionary.
       tune_filters: one of "block", "layer", "none" for tuning entire
         network, each layer separately, or no tuning.
       tune_filters_exceptions: name of layers that will not be tuned.
       layer_indexes: indexes of layer to be quantized.
       learning_rate_optimizer: if true, user will provide lr scheduler
         callback.
       blocks: list of re patterns specifygin group configuration for layers.
       schedule_block: "sequential" or "cost". Schedule blocks using the
         order of the groups or decreasing cost (energy or bits).
       quantization_config: file name of dictionary containing configuration of
         quantizers for kernel, bias and activation.
       debug: if True, fit will just print the groups for debugging purposes.
       head_name: specify which head to calcuate score/trial-size from in
         autoqkeras
       tuner_kwargs: parameters for keras_tuner depending on whether
         mode is random, hyperband or baeysian. Please refer to the
         documentation of kerstuner Tuners.
  """

  def __init__(
      self, model, metrics=None, custom_objects=None, goal=None,
      output_dir="result", mode="random", transfer_weights=False,
      activation_bits=4, limit=None, tune_filters="none",
      tune_filters_exceptions=None, layer_indexes=None,
      learning_rate_optimizer=False, blocks=None, schedule_block="sequential",
      quantization_config=None, overwrite=True, debug=False, head_name=None,
      **tuner_kwargs):

    if not metrics:
      metrics = []

    if not custom_objects:
      custom_objects = {}

    # goal: { "type": ["bits", "energy"], "params": {...} }
    # For type == "bits":
    #   delta_p: increment (in %) of the accuracy if trial is smaller.
    #   delta_n: decrement (in %) of the accuracy if trial is bigger.
    #   rate: rate of decrease/increase in model size in terms of bits.
    #   input_bits; size of input tensors.
    #   output_bits; size of output tensors.
    #   stress: parameter to reduce reference size to force tuner to
    #     choose smaller models.
    #   config: configuration on what to compute for each layer
    #     minimum configuration is { "default": ["parameters", "activations"] }

    # use simplest one - number of bits
    if not goal:
      goal = {
          "type": "bits",
          "params": {
              "delta_p": 8.0,
              "delta_n": 8.0,
              "rate": 2.0,
              "stress": 1.0,
              "input_bits": 8,
              "output_bits": 8,
              "ref_bits": 8,
              "config": {
                  "default": ["parameters", "activations"]
              }
          }
      }

    self.target = forgiving_factor[goal["type"]](**goal["params"])

    self.model = model
    self.metrics = metrics
    self.custom_objects = custom_objects
    self.mode = mode
    self.transfer_weights = transfer_weights
    self.activation_bits = activation_bits
    self.limit = limit
    self.tune_filters = tune_filters
    self.tune_filters_exceptions = tune_filters_exceptions
    self.layer_indexes = layer_indexes
    self.learning_rate_optimizer = learning_rate_optimizer
    self.blocks = blocks
    self.schedule_block = schedule_block
    self.quantization_config = quantization_config
    self.tuner_kwargs = tuner_kwargs
    self.debug = debug
    self.head_name = head_name

    self.autoqk = None
    self.learning_rate = model.optimizer.lr.numpy()
    self.overwrite = overwrite

    assert self.schedule_block in ["sequential", "cost"]

    # right now we create unique results directory
    idx = 0
    name = output_dir
    if self.overwrite:
      while os.path.exists(name):
        idx += 1
        name = output_dir + "_" + str(idx)
    output_dir = name
    self.output_dir = output_dir
    self.next_block = self.get_next_block(overwrite)
    if self.next_block > 0:
      strategy = self.tuner_kwargs.get("distribution_strategy", None)
      if strategy:
        with strategy.scope():
          self.model = tf.keras.models.load_model(
              os.path.join(
                  self.output_dir, "model_block_" + str(self.next_block - 1)),
              custom_objects=self.custom_objects)
      else:
        self.model = tf.keras.models.load_model(
            os.path.join(
                self.output_dir, "model_block_" + str(self.next_block - 1)),
            custom_objects=self.custom_objects)
      print("Load model completed")

  def get_next_block(self, overwrite):
    """Get the next block id to be worked on."""
    if overwrite:
      return 0
    else:
      try:
        with tf.io.gfile.GFile(os.path.join(self.output_dir, "scheduler.json"),
                               "r") as f:
          scheduler_json = f.read()
        scheduler = json.loads(scheduler_json)
        return scheduler["next_block"]
      except:  # pylint: disable=bare-except
        return 0

  def get_limit(self, model, pattern):
    """Apply patterned group to limit to obtain new limit set."""
    limit = self.limit
    new_limit = {}
    new_pattern = collections.defaultdict(list)

    for layer_name in self.grouped_patterns[pattern]:
      layer = model.get_layer(layer_name)
      layer_class_name = layer.__class__.__name__

      target_quantizers = limit.get(layer_class_name, -1)
      for limit_pattern in limit:
        if re.match(limit_pattern, layer_name):
          target_quantizers = limit[limit_pattern]
          new_pattern[limit_pattern].append(layer_name)
          layer_name = limit_pattern
          break
      if target_quantizers != -1:
        new_limit[layer_name] = target_quantizers

    for key in new_pattern:
      # grouped pattern in regex need to be ^(word1|word2|...)$ instead of
      # ^word1|word2|...$; otherwise it cause non-exact match,
      # e.g., fc.*_0 and fc.*_0_relu were miss-matched
      new_key = "^" + "(" + "|".join(new_pattern[key]) + ")" + "$"
      new_limit[new_key] = new_limit[key]
      if new_key != key:
        del new_limit[key]

    return new_limit

  def fit(self, *fit_args, **fit_kwargs):
    """Invokes tuner fit algorithm."""

    self.history = []
    self.compute_block_costs(self.blocks, self.model)

    if self.tuner_kwargs.get("max_trials", None):
      max_trials = float(self.tuner_kwargs["max_trials"])

    lr = self.model.optimizer.lr.numpy()

    model = self.model

    frozen_layers = []

    for i, (pattern, cost) in enumerate(self.retrieve_max_block()):

      # now create new limit pattern
      if not self.overwrite:
        if i < self.next_block:
          print("Resume tuning. Skipping block ", i)
          continue

      print("... block cost: {:.0f} / {:.0f}".format(cost, self.reference_size))

      if self.tuner_kwargs.get("max_trials", None):
        self.tuner_kwargs["max_trials"] = int(
            max(10, max_trials * cost / self.reference_size))
        print("... adjusting max_trials for this block to {}".format(
            self.tuner_kwargs["max_trials"]))

      limit = self.get_limit(model, pattern)
      new_frozen_layers = self.grouped_patterns[pattern]

      # if dictionary is empty we did not match anything.
      # we have a bug in the patterns specified by the
      # user.

      assert limit

      print("Pattern {} is : {}".format(i, limit))

      if self.debug:
        frozen_layers = frozen_layers + new_frozen_layers
        continue

      self.autoqk = AutoQKeras(
          model, self.metrics,
          custom_objects=self.custom_objects,
          goal=self.target,
          output_dir=self.output_dir + "/" + str(i),
          mode=self.mode,
          transfer_weights=self.transfer_weights,
          frozen_layers=frozen_layers,
          activation_bits=self.activation_bits,
          limit=limit,
          tune_filters=self.tune_filters,
          tune_filters_exceptions=self.tune_filters_exceptions,
          layer_indexes=self.layer_indexes,
          learning_rate_optimizer=self.learning_rate_optimizer,
          quantization_config=self.quantization_config,
          overwrite=self.overwrite,
          head_name=self.head_name,
          **self.tuner_kwargs)

      self.autoqk.fit(*fit_args, **fit_kwargs)

      self.autoqk.tuner.results_summary()

      self.history.append(self.autoqk.history())

      model = self.autoqk.get_best_model()
      self.learning_rate = model.optimizer.lr.numpy()

      # restore learning rate
      # this is just a placeholder for the optimizer.

      model.compile(
          model.optimizer,
          loss=self.model.loss,
          metrics=self.model.metrics)

      frozen_layers = frozen_layers + new_frozen_layers

      filename = self.output_dir + "/model_block_" + str(i)
      model.save(filename)
      self.next_block = i + 1

      # update scheduler json
      with tf.io.gfile.GFile(os.path.join(self.output_dir, "scheduler.json"),
                             "w") as f:
        f.write(json.dumps({"next_block": self.next_block}))

    if self.debug:
      return

    self.best_model = model

    # make all layers trainable again
    for layer_name in frozen_layers:
      layer = model.get_layer(layer_name)
      layer.trainable = True

  def compute_block_costs(self, patterns, model):
    """Computes costs for each block."""

    # get block cost for original model
    self.reference_size = self.target.get_reference(model)
    self.model_size = self.target.get_reference_stats()

    # first group layers into the patterns

    groups = {pattern: [] for pattern in patterns}

    for layer_id, layer in enumerate(model.layers):
      if (
          self.layer_indexes is not None and
          layer_id not in self.layer_indexes
      ):
        continue

      for pattern in groups:
        if re.match(pattern, layer.name):
          groups[pattern].append(layer.name)

    self.grouped_patterns = groups

    # now compute cost for each group

    self.costs = []
    for pattern in patterns:  # self.grouped_patterns:
      total = 0
      for layer in self.grouped_patterns[pattern]:
        if layer in self.model_size:
          total += self.model_size[layer]["total"]
      self.costs.append((pattern, total))

    # the costs will be sorted by the total cost of the group
    if self.schedule_block == "cost":
      self.costs = sorted(self.costs, key=lambda cost_tuple: -cost_tuple[1])

  def retrieve_max_block(self):
    for cost in self.costs:
      yield cost

  def get_history(self):
    """Returns the history of the model search."""
    return self.history

  def get_best_model(self):
    """Returns the best model."""

    # check if we have run fit first.
    if not self.autoqk:
      return None

    self.autoqk.hypermodel.target.print_stats()
    print_qmodel_summary(self.best_model)

    return self.best_model

  def get_learning_rate(self):
    return self.learning_rate



# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements a safe evaluation using globals()."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyparsing import delimitedList
from pyparsing import Group
from pyparsing import Optional
from pyparsing import Regex
from pyparsing import Suppress

import logging
from tensorflow import keras


def Num(s):
  """Tries to convert string to either int or float."""
  try:
    try:
      return int(s)
    except ValueError:
      return float(s)
  except ValueError:
    # this should be always true. if it isn't int or float, it should be str
    assert (
        (s[0] == '"' and s[-1] == '"') or
        (s[0] == "'" and s[-1] == "'")
    )
    s = s[1:-1]
    return s

def Str(s):
  return s[1:-1]

def IsNum(s):
  try:
    try:
      int(s)
      return True
    except ValueError:
      float(s)
      return True
  except ValueError:
    return False

def IsBool(s):
  if s in ["True", "False"]:
    return True
  else:
    return False

def Bool(s):
  return True if "True" in s else False

def GetArg(s):
  if IsBool(s):
    return Bool(s)
  elif IsNum(s):
    return Num(s)
  else:
    return Str(s)


def GetParams(s):
  """Extracts args and kwargs from string."""
  # modified from https://stackoverflow.com/questions/38799223/parse-string-to-identify-kwargs-and-args  # pylint: disable=line-too-long

  _lparen = Suppress("(")  # pylint: disable=invalid-name
  _rparen = Suppress(")")  # pylint: disable=invalid-name
  _eq = Suppress("=")  # pylint: disable=invalid-name

  data = (_lparen + Optional(
      delimitedList(
          Group(Regex(r"[^=,)\s]+") + Optional(_eq + Regex(u"[^,)]*")))
          )
      ) + _rparen)

  items = data.parseString(s).asList()

  # need to make sure that kwargs only happen after args are processed
  args = [GetArg(i[0]) for i in items if len(i) == 1]
  kwargs = {i[0]: GetArg(i[1]) for i in items if len(i) == 2}

  # check for syntax error
  for i in range(1, len(items)):
    if (len(items[i]) == 1) and (len(items[i-1]) == 2):
      raise SyntaxError

  return args, kwargs


def safe_eval(eval_str, op_dict, *params, **kwparams):  # pylint: disable=invalid-name
  """Replaces eval by a safe eval mechanism."""

  function_split = eval_str.split("(")
  quantizer = op_dict.get(function_split[0], None)

  if len(function_split) == 2:
    args, kwargs = GetParams("(" + function_split[1])
  else:
    args = []
    kwargs = {}

  args = args + list(params)
  for k in kwparams:
    kwargs[k] = kwparams[k]

  # must be Keras activation object if None
  if quantizer is None:
    logging.info("keras dict %s", function_split[0])
    quantizer = keras.activations.get(function_split[0])

  if len(function_split) == 2 or args or kwargs:
    return quantizer(*args, **kwargs)
  else:
    if isinstance(quantizer, type):
      # Check if quantizer is a class
      return quantizer()
    else:
      # Otherwise it is a function, so just return it
      return quantizer



# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Exports qkeras modules to quantizer package."""

# We use wildcard import for convenience at this moment, which will be later
# refactored and removed.
import tensorflow as tf

from .b2t import *  # pylint: disable=wildcard-import
from .estimate import *  # pylint: disable=wildcard-import
from .qlayers import *  # pylint: disable=wildcard-import
from .quantizers import *  # pylint: disable=wildcard-import
from .qconvolutional import *  # pylint: disable=wildcard-import
from .qrecurrent import *  # pylint: disable=wildcard-import
from .qnormalization import * # pylint: disable=wildcard-import
from .qoctave import *  # pylint: disable=wildcard-import
from .qpooling import *  # pylint: disable=wildcard-import
from .safe_eval import *  # pylint: disable=wildcard-import
#from .qtools.run_qtools import QTools
#from .qtools.settings import cfg
from .qconv2d_batchnorm import QConv2DBatchnorm
from .qdepthwiseconv2d_batchnorm import QDepthwiseConv2DBatchnorm

assert tf.executing_eagerly(), "QKeras requires TF with eager execution mode on"

__version__ = "0.9.0"



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""implement adder quantizer."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

from absl import logging
from qkeras.qtools.quantized_operators import adder_impl
from qkeras.qtools.quantized_operators import quantizer_impl


class IAdder(abc.ABC):
  """abstract class for adder."""

  def __init__(self):
    self.adder_impl_table = [
        [
            adder_impl.FixedPointAdder,
            adder_impl.Po2FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FloatingPointAdder
        ],
        [
            adder_impl.Po2FixedPointAdder,
            adder_impl.Po2Adder,
            adder_impl.Po2FixedPointAdder,
            adder_impl.Po2FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FloatingPointAdder
        ],
        [
            adder_impl.FixedPointAdder,
            adder_impl.Po2FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FloatingPointAdder
        ],
        [
            adder_impl.FixedPointAdder,
            adder_impl.Po2FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FloatingPointAdder
        ],
        [
            adder_impl.FixedPointAdder,
            adder_impl.Po2FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FixedPointAdder,
            adder_impl.FloatingPointAdder
        ],
        [
            adder_impl.FloatingPointAdder,
            adder_impl.FloatingPointAdder,
            adder_impl.FloatingPointAdder,
            adder_impl.FloatingPointAdder,
            adder_impl.FloatingPointAdder,
            adder_impl.FloatingPointAdder
        ]
    ]

  def make_quantizer(self, quantizer_1: quantizer_impl.IQuantizer,
                     quantizer_2: quantizer_impl.IQuantizer):
    """make adder quantizer."""

    self.quantizer_1 = quantizer_1
    self.quantizer_2 = quantizer_2

    mode1 = quantizer_1.mode
    mode2 = quantizer_2.mode

    adder_impl_class = self.adder_impl_table[mode1][mode2]
    logging.debug(
        "qbn adder implemented as class %s",
        adder_impl_class.implemented_as())

    return adder_impl_class(
        quantizer_1,
        quantizer_2
    )



# ==============================================================================
# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Definition of default quantization configuration."""

default_quantization_config = {
    "kernel": {
        "binary": 1,
        "stochastic_binary": 1,
        "ternary": 2,
        "stochastic_ternary": 2,
        "quantized_bits(2,1,1,alpha=1.0)": 2,
        "quantized_bits(4,0,1)": 4,
        "quantized_bits(8,0,1)": 8,
        "quantized_po2(4,1)": 4
    },
    "bias": {
        "quantized_bits(4,0,1)": 4,
        "quantized_bits(8,3,1)": 8,
        "quantized_po2(4,8)": 4
    },
    "activation": {
        "binary": 1,
        "binary(alpha='auto_po2')": 1,
        "ternary": 2,
        "quantized_relu(3,1)": 3,
        "quantized_relu(4,2)": 4,
        "quantized_relu(8,2)": 8,
        "quantized_relu(8,4)": 8,
        "quantized_relu(16,8)": 16,
        "quantized_relu_po2(4,4)": 4
    },
    "linear": {
        "binary": 1,
        "ternary": 2,
        "quantized_bits(4,1)": 4,
        "quantized_bits(8,2)": 8,
        "quantized_bits(16,10)": 16,
        "quantized_po2(6,4)": 6
    }
}



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Create accumulator quantizers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from qkeras.qtools.quantized_operators import accumulator_impl
from qkeras.qtools.quantized_operators import multiplier_impl


class AccumulatorFactory:
  """interface for accumulator type."""

  def make_accumulator(
      self, kernel_shape,
      multiplier: multiplier_impl.IMultiplier,
      use_bias=True
  ) -> accumulator_impl.IAccumulator:
    """Create an accumulator instance."""

    # The type and bit width of the accumulator is deteremined from the
    # multiplier implementation, and the shape of both kernel and bias

    if multiplier.output.is_floating_point:
      accumulator = accumulator_impl.FloatingPointAccumulator(
          multiplier)

    # po2*po2 is implemented as Adder; output type is po2
    # in multiplier, po2 needs to be converted to FixedPoint
    elif multiplier.output.is_po2:
      accumulator = accumulator_impl.Po2Accumulator(
          kernel_shape, multiplier, use_bias)

    # fixed point
    else:
      accumulator = accumulator_impl.FixedPointAccumulator(
          kernel_shape, multiplier, use_bias)

    return accumulator



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Accumulator operation implementation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
from absl import logging
import numpy as np

from qkeras.qtools.quantized_operators import multiplier_impl
from qkeras.qtools.quantized_operators import quantizer_impl


def po2_to_qbits(quantizer: quantizer_impl.IQuantizer):
  """convert po2 type to qbits type."""

  (min_exp, max_exp) = quantizer.get_min_max_exp()
  # min_exp is number of bits needed on the right in qbits
  # max_exp is number of bits needed on the left in qbits
  unsigned_bits = min_exp + max_exp
  int_bits = max_exp
  sign_bit = quantizer.is_signed
  bits = sign_bit + unsigned_bits

  return (int(bits), int(int_bits))


class IAccumulator(abc.ABC):
  """abstract class for accumulator."""

  @staticmethod
  @abc.abstractmethod
  def implemented_as():
    pass


class FloatingPointAccumulator(IAccumulator):
  """class for floating point accumulator."""

  def __init__(
      self,
      multiplier: multiplier_impl.IMultiplier
  ):
    super().__init__()

    self.multiplier = multiplier
    self.output = quantizer_impl.FloatingPoint(
        bits=self.multiplier.output.bits)
    self.output.bits = self.multiplier.output.bits
    self.output.int_bits = -1
    self.output.is_signed = self.multiplier.output.is_signed
    self.output.is_floating_point = True
    self.output.op_type = "accumulator"

  @staticmethod
  def implemented_as():
    return "add"


class FixedPointAccumulator(IAccumulator):
  """class for fixed point accumulator."""

  def __init__(
      self,
      kernel_shape,
      multiplier: multiplier_impl.IMultiplier,
      use_bias=True
  ):
    super().__init__()

    if len(kernel_shape) not in (
        2,
        4,
    ):
      logging.fatal(
          "unsupported kernel shape, "
          "it is neither a dense kernel of length 2,"
          " nor a convolution kernel of length 4")

    kernel_shape_excluding_output_dim = kernel_shape[:-1]
    kernel_add_ops = np.prod(kernel_shape_excluding_output_dim)

    # bias are associate with filters; each filter adds 1 bias
    bias_add = 1 if use_bias else 0

    add_ops = kernel_add_ops + bias_add
    self.log_add_ops = int(np.ceil(np.log2(add_ops)))

    self.multiplier = multiplier
    self.output = quantizer_impl.QuantizedBits()
    self.output.bits = self.log_add_ops + self.multiplier.output.bits
    self.output.int_bits = self.log_add_ops + self.multiplier.output.int_bits
    self.output.is_signed = self.multiplier.output.is_signed
    self.output.op_type = "accumulator"

    assert not self.multiplier.output.is_floating_point
    self.output.is_floating_point = False

  @staticmethod
  def implemented_as():
    return "add"


class Po2Accumulator(FixedPointAccumulator):
  """accumulator for po2."""

  # multiplier is po2. multiplier output needs to convert
  # to Fixedpoint before Accumulator.

  def __init__(
      self,
      kernel_shape,
      multiplier: multiplier_impl.IMultiplier,
      use_bias=True
  ):
    super().__init__(kernel_shape, multiplier, use_bias)

    assert multiplier.output.is_po2
    # convert multiplier output from po2 to quantized_bits
    (bits_from_po2multiplier, int_bits_from_po2multiplier) = po2_to_qbits(
        multiplier.output)

    self.output.bits = self.log_add_ops + int(bits_from_po2multiplier)
    self.output.int_bits = self.log_add_ops + int(int_bits_from_po2multiplier)
    self.output.op_type = "accumulator"

  @staticmethod
  def implemented_as():
    return "add"



# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Octave Convolution."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re

from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Add
from tensorflow.keras.layers import AveragePooling2D
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import SeparableConv2D
from tensorflow.keras.layers import UpSampling2D
from .qlayers import QActivation
from .qconvolutional import QConv2D
from .qconvolutional import QSeparableConv2D
from .qpooling import QAveragePooling2D


def GetActivationSuffix(activation):
  """Returns suffix for layer name to facilitate debugging."""
  if not activation:
    return "linear"

  if "po2" in activation:
    return "q2"
  elif "quantized_relu" in activation:
    suffix = "qr"
  elif "quantized_tanh" in activation:
    suffix = "qt"
  else:
    suffix = "qb"

  numbers = re.findall(r"[0-9]+", activation)

  numbers = [n + "_" if len(n) > 1 else n for n in numbers]

  return suffix + "".join(numbers)


def QOctaveConv2D(
    filters,
    kernel_size,
    alpha,
    strides=(1, 1),
    padding="valid",
    kernel_initializer="he_normal",
    bias_initializer="zeros",
    # NOTE: kernel_regularizer not used with separable convolution
    kernel_regularizer=None,
    bias_regularizer=None,
    kernel_constraint=None,
    bias_constraint=None,
    use_separable=True,
    name="",
    **kwargs):
  """Implements quantized QOctaveConv2D."""

  def _QOctaveConv2DInternal(x):
    """Computes QOctaveConv2D on a tensor."""

    x_h, x_l = x

    bias_quantizer = kwargs.get("bias_quantizer", None)
    kernel_quantizer = kwargs.get("kernel_quantizer", None)
    depthwise_quantizer = kwargs.get("depthwise_quantizer", None)
    pointwise_quantizer = kwargs.get("pointwise_quantizer", None)
    acc_quantizer = kwargs.get("acc_quantizer", None)
    pooling_quantizer = kwargs.get("pooling_quantizer", None)
    depthwise_activation = kwargs.get("depthwise_activation", None)
    activation = kwargs.get("activation", None)

    bias_range = kwargs.get("bias_range", 1.0)
    kernel_range = kwargs.get("kernel_range", 1.0)
    depthwise_range = kwargs.get("depthwise_range", 1.0)
    pointwise_range = kwargs.get("pointwise_range", 1.0)

    if activation:
      act_suffix = "_" + GetActivationSuffix(activation)
    acc_suffix = "_" + GetActivationSuffix(acc_quantizer)

    if alpha == -1.0:
      if use_separable:
        x_h = QSeparableConv2D(
            filters, kernel_size, strides=strides, padding=padding,
            depthwise_regularizer=kernel_regularizer,
            depthwise_constraint=kernel_constraint,
            depthwise_initializer=kernel_initializer,
            pointwise_regularizer=kernel_regularizer,
            pointwise_constraint=kernel_constraint,
            pointwise_initializer=kernel_initializer,
            bias_regularizer=bias_regularizer,
            bias_constraint=bias_constraint,
            bias_initializer=bias_initializer,
            depthwise_quantizer=depthwise_quantizer,
            pointwise_quantizer=pointwise_quantizer,
            bias_quantizer=bias_quantizer,
            depthwise_activation=depthwise_activation,
            pointwise_range=pointwise_range,
            depthwise_range=depthwise_range,
            bias_range=bias_range,
            name=name + "_c_h_to_h")(x_h)
      else:
        x_h = QConv2D(
            filters, kernel_size, strides=strides, padding=padding,
            kernel_regularizer=kernel_regularizer,
            kernel_constraint=kernel_constraint,
            kernel_initializer=kernel_initializer,
            bias_regularizer=bias_regularizer,
            bias_constraint=bias_constraint,
            bias_initializer=bias_initializer,
            kernel_quantizer=kernel_quantizer,
            bias_quantizer=bias_quantizer,
            kernel_range=kernel_range,
            bias_range=bias_range,
            name=name + "_c_h_to_h")(x_h)

      if activation:
        x_h = QActivation(
            activation, name=name + "_c_h_to_h_act" + act_suffix)(
                x_h)

      return [x_h, None]

    co_h = int(filters * (1 - alpha))
    co_l = filters - co_h

    x_h_to_h = None
    x_h_to_l = None
    x_l_to_l = None
    x_l_to_h = None

    if co_h > 0:
      if x_h is not None:
        if use_separable:
          x_h_to_h = QSeparableConv2D(
              co_h, kernel_size, strides=strides, padding=padding,
              depthwise_regularizer=kernel_regularizer,
              depthwise_constraint=kernel_constraint,
              depthwise_initializer=kernel_initializer,
              pointwise_regularizer=kernel_regularizer,
              pointwise_constraint=kernel_constraint,
              pointwise_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              depthwise_quantizer=depthwise_quantizer,
              pointwise_quantizer=pointwise_quantizer,
              bias_quantizer=bias_quantizer,
              depthwise_activation=depthwise_activation,
              pointwise_range=pointwise_range,
              depthwise_range=depthwise_range,
              bias_range=bias_range,
              name=name + "_c_h_to_h")(x_h)
        else:
          x_h_to_h = QConv2D(
              co_h, kernel_size, strides=strides, padding=padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              kernel_quantizer=kernel_quantizer,
              bias_quantizer=bias_quantizer,
              kernel_range=kernel_range,
              bias_range=bias_range,
              name=name + "_c_h_to_h")(x_h)

        if acc_quantizer:
          x_h_to_h = QActivation(
              acc_quantizer,
              name=name + "_c_h_to_h_act" + acc_suffix)(x_h_to_h)

    if co_l > 0:
      if x_h is not None:
        x_h_to_l = QAveragePooling2D(
            pool_size=2, strides=2,
            quantizer=pooling_quantizer,
            name=name + "_avg_h_to_l")(x_h)

        if use_separable:
          x_h_to_l = QSeparableConv2D(
              co_l, kernel_size, strides=strides, padding=padding,
              depthwise_regularizer=kernel_regularizer,
              depthwise_constraint=kernel_constraint,
              depthwise_initializer=kernel_initializer,
              pointwise_regularizer=kernel_regularizer,
              pointwise_constraint=kernel_constraint,
              pointwise_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              depthwise_quantizer=depthwise_quantizer,
              pointwise_quantizer=pointwise_quantizer,
              bias_quantizer=bias_quantizer,
              depthwise_activation=depthwise_activation,
              pointwise_range=pointwise_range,
              depthwise_range=depthwise_range,
              bias_range=bias_range,
              name=name + "_c_h_to_l")(x_h_to_l)
        else:
          x_h_to_l = QConv2D(
              co_l, kernel_size, strides=strides, padding=padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              kernel_quantizer=kernel_quantizer,
              bias_quantizer=bias_quantizer,
              kernel_range=kernel_range,
              bias_range=bias_range,
              name=name + "_c_h_to_l")(x_h_to_l)

        if acc_quantizer:
          x_h_to_l = QActivation(
              acc_quantizer,
              name=name + "_c_h_to_l_act" + acc_suffix)(x_h_to_l)

    if co_h > 0:
      if x_l is not None:
        _, height, width, _ = x_l.shape.as_list()
        if height == 1 and width == 1:
          local_kernel = 1
          local_strides = 1
          local_padding = "same"
          upsampling = False
        else:
          local_kernel = kernel_size
          local_strides = strides
          local_padding = padding
          upsampling = True

        if use_separable and upsampling:
          x_l_to_h = QSeparableConv2D(
              co_h, kernel_size, strides=strides, padding=padding,
              depthwise_regularizer=kernel_regularizer,
              depthwise_constraint=kernel_constraint,
              depthwise_initializer=kernel_initializer,
              pointwise_regularizer=kernel_regularizer,
              pointwise_constraint=kernel_constraint,
              pointwise_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              depthwise_quantizer=depthwise_quantizer,
              pointwise_quantizer=pointwise_quantizer,
              bias_quantizer=bias_quantizer,
              depthwise_activation=depthwise_activation,
              pointwise_range=pointwise_range,
              depthwise_range=depthwise_range,
              bias_range=bias_range,
              name=name + "_c_l_to_h")(x_l)
        else:
          x_l_to_h = QConv2D(
              co_h, local_kernel, strides=local_strides, padding=local_padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              kernel_quantizer=kernel_quantizer,
              bias_quantizer=bias_quantizer,
              kernel_range=kernel_range,
              bias_range=bias_range,
              name=name + "_c_l_to_h")(x_l)

        if acc_quantizer:
          x_l_to_h = QActivation(
              acc_quantizer,
              name=name + "_c_l_to_h_act" + acc_suffix)(x_l_to_h)

        if upsampling:
          x_l_to_h = UpSampling2D(
              size=(2, 2), name=name + "_u_l_to_h")(x_l_to_h)

    if co_l > 0:
      if x_l is not None:
        if use_separable:
          x_l_to_l = QSeparableConv2D(
              co_l, kernel_size, strides=strides, padding=padding,
              depthwise_regularizer=kernel_regularizer,
              depthwise_constraint=kernel_constraint,
              depthwise_initializer=kernel_initializer,
              pointwise_regularizer=kernel_regularizer,
              pointwise_constraint=kernel_constraint,
              pointwise_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              depthwise_quantizer=depthwise_quantizer,
              pointwise_quantizer=depthwise_quantizer,
              bias_quantizer=bias_quantizer,
              depthwise_activation=depthwise_activation,
              pointwise_range=pointwise_range,
              depthwise_range=depthwise_range,
              bias_range=bias_range,
              name=name + "_c_l_to_l")(x_l)
        else:
          x_l_to_l = QConv2D(
              co_l, kernel_size, strides=strides, padding=padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              kernel_quantizer=kernel_quantizer,
              bias_quantizer=bias_quantizer,
              kernel_range=kernel_range,
              bias_range=bias_range,
              name=name + "_c_l_to_l")(x_l)

        if acc_quantizer:
          x_l_to_l = QActivation(
              acc_quantizer, name=name + "_c_l_to_l_act" + acc_suffix)(
                  x_l_to_l)

    if x_h_to_h is not None and x_l_to_h is not None:
      x_h = Add(name=name + "_a_h")([x_h_to_h, x_l_to_h])
    elif x_h_to_h is not None:
      x_h = x_h_to_h
    elif x_l_to_h is not None:
      x_h = x_l_to_h
    else:
      x_h = None

    if x_l_to_l is not None and x_h_to_l is not None:
      x_l = Add(name=name + "_a_l")([x_l_to_l, x_h_to_l])
    elif x_l_to_l is not None:
      x_l = x_l_to_l
    elif x_h_to_l is not None:
      x_l = x_h_to_l
    else:
      x_l = None

    if x_h is not None and activation is not None:
      x_h = QActivation(activation,
                        name=name + "_h_act" + act_suffix)(x_h)

    if x_l is not None and activation is not None:
      x_l = QActivation(activation,
                        name=name + "_l_act" + act_suffix)(x_l)

    return [x_h, x_l]

  return _QOctaveConv2DInternal


def OctaveConv2D(
    filters, kernel_size, alpha,
    strides=(1, 1), padding="valid",
    kernel_initializer="he_normal",
    bias_initializer="zeros",
    kernel_regularizer=None,
    bias_regularizer=None,
    kernel_constraint=None,
    bias_constraint=None,
    activation=None,
    use_separable=True,
    name="",
    **kwargs):

  """Implements OctaveConv2D."""

  def _OctaveConv2DInternal(x):

    """Computes octave on tensor."""

    acc_quantizer = kwargs.get("acc_quantizer", None)

    x_h, x_l = x

    if alpha == -1.0:
      if use_separable:
        x_h = SeparableConv2D(
            filters, kernel_size, strides=strides, padding=padding,
            depthwise_regularizer=kernel_regularizer,
            depthwise_constraint=kernel_constraint,
            depthwise_initializer=kernel_initializer,
            pointwise_regularizer=kernel_regularizer,
            pointwise_constraint=kernel_constraint,
            pointwise_initializer=kernel_initializer,
            bias_regularizer=bias_regularizer,
            bias_constraint=bias_constraint,
            bias_initializer=bias_initializer,
            name=name + "_c_h_to_h")(x_h)
      else:
        x_h = Conv2D(
            filters, kernel_size, strides=strides, padding=padding,
            kernel_regularizer=kernel_regularizer,
            kernel_constraint=kernel_constraint,
            kernel_initializer=kernel_initializer,
            bias_regularizer=bias_regularizer,
            bias_constraint=bias_constraint,
            bias_initializer=bias_initializer,
            name=name+"_c_h_to_h")(x_h)

      if activation:
        x_h = Activation(activation, name=name + "_c_h_to_h_act")(x_h)

      return [x_h, None]

    co_h = int(filters * (1 - alpha))
    co_l = filters - co_h

    x_h_to_h = None
    x_h_to_l = None
    x_l_to_l = None
    x_l_to_h = None

    if co_h > 0:
      if x_h is not None:
        if use_separable:
          x_h_to_h = SeparableConv2D(
              co_h, kernel_size, strides=strides, padding=padding,
              depthwise_regularizer=kernel_regularizer,
              depthwise_constraint=kernel_constraint,
              depthwise_initializer=kernel_initializer,
              pointwise_regularizer=kernel_regularizer,
              pointwise_constraint=kernel_constraint,
              pointwise_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              name=name + "_c_h_to_h")(x_h)
        else:
          x_h_to_h = Conv2D(
              co_h, kernel_size, strides=strides, padding=padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              name=name + "_c_h_to_h")(x_h)

        if activation:
          x_h_to_h = Activation(
              acc_quantizer, name=name + "_c_h_to_h_act")(x_h_to_h)

    if co_l > 0:
      if x_h is not None:
        x_h_to_l = AveragePooling2D(pool_size=2, strides=2,
                                    name=name + "_p_h_to_l")(x_h)

        if use_separable:
          x_h_to_l = SeparableConv2D(
              co_l, kernel_size, strides=strides, padding=padding,
              depthwise_regularizer=kernel_regularizer,
              depthwise_constraint=kernel_constraint,
              depthwise_initializer=kernel_initializer,
              pointwise_regularizer=kernel_regularizer,
              pointwise_constraint=kernel_constraint,
              pointwise_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              name=name + "_c_h_to_l")(x_h_to_l)
        else:
          x_h_to_l = Conv2D(
              co_l, kernel_size, strides=strides, padding=padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              name=name + "_c_h_to_l")(x_h_to_l)

        if activation:
          x_h_to_l = Activation(
              acc_quantizer, name=name + "_c_h_to_l_act")(x_h_to_l)

    if co_h > 0:
      if x_l is not None:
        _, height, width, _ = x_l.shape.as_list()
        if height == 1 and width == 1:
          local_kernel = 1
          local_strides = 1
          local_padding = "same"
          upsampling = False
        else:
          local_kernel = kernel_size
          local_strides = strides
          local_padding = padding
          upsampling = True

        if use_separable and upsampling:
          x_l_to_h = SeparableConv2D(
              co_h, kernel_size, strides=strides, padding=padding,
              depthwise_regularizer=kernel_regularizer,
              depthwise_constraint=kernel_constraint,
              depthwise_initializer=kernel_initializer,
              pointwise_regularizer=kernel_regularizer,
              pointwise_constraint=kernel_constraint,
              pointwise_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              name=name + "_c_l_to_h")(x_l)
        else:
          x_l_to_h = Conv2D(
              co_h, local_kernel, strides=local_strides, padding=local_padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              name=name + "_c_l_to_h")(x_l)

        if activation:
          x_l_to_h = Activation(
              acc_quantizer, name=name + "_c_l_to_h_act")(x_l_to_h)

        if upsampling:
          x_l_to_h = UpSampling2D(
              size=(2, 2), name=name + "_u_l_to_h")(x_l_to_h)

    if co_l > 0:
      if x_l is not None:
        if use_separable:
          x_l_to_l = SeparableConv2D(
              co_l, kernel_size, strides=strides, padding=padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              name=name + "_c_l_to_l")(x_l)
        else:
          x_l_to_l = Conv2D(
              co_l, kernel_size, strides=strides, padding=padding,
              kernel_regularizer=kernel_regularizer,
              kernel_constraint=kernel_constraint,
              kernel_initializer=kernel_initializer,
              bias_regularizer=bias_regularizer,
              bias_constraint=bias_constraint,
              bias_initializer=bias_initializer,
              name=name + "_c_l_to_l")(x_l)

        if activation:
          x_l_to_l = Activation(
              acc_quantizer, name=name + "_c_l_to_l_act")(x_l_to_l)

    if x_h_to_h is not None and x_l_to_h is not None:
      x_h = Add(name=name + "_a_h")([x_h_to_h, x_l_to_h])
    elif x_h_to_h is not None:
      x_h = x_h_to_h
    elif x_l_to_h is not None:
      x_h = x_l_to_h
    else:
      x_h = None

    if x_l_to_l is not None and x_h_to_l is not None:
      x_l = Add(name=name + "_a_l")([x_l_to_l, x_h_to_l])
    elif x_l_to_l is not None:
      x_l = x_l_to_l
    elif x_h_to_l is not None:
      x_l = x_h_to_l
    else:
      x_l = None

    if x_h is not None:
      x_h = Activation(activation, name=name + "_h_act")(x_h)

    if x_l is not None:
      x_l = Activation(activation, name=name + "_l_act")(x_l)

    return (x_h, x_l)

  return _OctaveConv2DInternal
